from zipfile import ZipFile
import os
import urllib
import urllib.request
import cv2
import pickle
import copy


"""

URL = 'https://nnfs.io/datasets/fashion_mnist_images.zip'
FILE = 'fashion_mnist_images.zip'
FOLDER = 'fashion_mnist_images'

if not os.path.isfile(FILE):
    print ( 'Downloading {URL} and saving as {FILE} ...' )
    urllib.request.urlretrieve(URL, FILE)

print ( 'Unzipping images...' )

with ZipFile(FILE) as zip_images:
    zip_images.extractall(FOLDER)

"""






import matplotlib.pyplot as plt


#Full code



import numpy as np
import nnfs
from nnfs.datasets import sine_data, spiral_data


nnfs.init()


# Dense layer
class Layer_Dense :
    # Layer initialization
    def __init__ ( self , n_inputs , n_neurons ,
    weight_regularizer_l1 = 0 , weight_regularizer_l2 = 0 ,
    bias_regularizer_l1 = 0 , bias_regularizer_l2 = 0 ):
        # Initialize weights and biases
        self.weights = 0.01 * np.random.randn(n_inputs, n_neurons)
        self.biases = np.zeros(( 1 , n_neurons))
        # Set regularization strength
        self.weight_regularizer_l1 = weight_regularizer_l1
        self.weight_regularizer_l2 = weight_regularizer_l2
        self.bias_regularizer_l1 = bias_regularizer_l1
        self.bias_regularizer_l2 = bias_regularizer_l2
    # Forward pass
    def forward ( self , inputs , training ):
        # Remember input values
        self.inputs = inputs
        # Calculate output values from inputs, weights and biases
        self.output = np.dot(inputs, self.weights) + self.biases
    # Backward pass
    def backward ( self , dvalues ):
        # Gradients on parameters
        self.dweights = np.dot(self.inputs.T, dvalues)
        self.dbiases = np.sum(dvalues, axis = 0 , keepdims = True )
        # Gradients on regularization
        # L1 on weights
        if self.weight_regularizer_l1 > 0 :
            dL1 = np.ones_like(self.weights)
            dL1[self.weights < 0 ] = - 1
            self.dweights += self.weight_regularizer_l1 * dL1
        # L2 on weights
        if self.weight_regularizer_l2 > 0 :
            self.dweights += 2 * self.weight_regularizer_l2 * self.weights
        # L1 on biases
        if self.bias_regularizer_l1 > 0 :
            dL1 = np.ones_like(self.biases)
            dL1[self.biases < 0 ] = - 1
            self.dbiases += self.bias_regularizer_l1 * dL1

        # L2 on biases
        if self.bias_regularizer_l2 > 0 :
            self.dbiases += 2 * self.bias_regularizer_l2 * self.biases

        # Gradient on values
        self.dinputs = np.dot(dvalues, self.weights.T)

        # Retrieve layer parameters
    def get_parameters(self):
            return self.weights, self.biases

        # Set weights and biases in a layer instance
    def set_parameters(self, weights, biases):
            self.weights = weights

            self.biases = biases



# Dropout
class Layer_Dropout :

    # Init
    def __init__ ( self , rate ):
        # Store rate, we invert it as for example for dropout
        # of 0.1 we need success rate of 0.9
        self.rate = 1 - rate

    # Forward pass
    def forward ( self , inputs , training ):
        # Save input values
        self.inputs = inputs
        # If not in the training mode - return values
        if not training:
            self.output = inputs.copy()
            return
        # Generate and save scaled mask
        self.binary_mask = np.random.binomial( 1 , self.rate,
        size = inputs.shape) / self.rate
        # Apply mask to output values
        self.output = inputs * self.binary_mask

    # Backward pass
    def backward ( self , dvalues ):
        # Gradient on values
        self.dinputs = dvalues * self.binary_mask
        # Input "layer"


class Layer_Input :
    # Forward pass
    def forward ( self , inputs , training ):
        self.output = inputs


# ReLU activation
class Activation_ReLU :
    # Forward pass
    def forward ( self , inputs , training ):
        # Remember input values
        self.inputs = inputs
        # Calculate output values from inputs
        self.output = np.maximum( 0 , inputs)
    # Backward pass
    def backward ( self , dvalues ):
        # Since we need to modify original variable,
        # let's make a copy of values first
        self.dinputs = dvalues.copy()
        # Zero gradient where input values were negative
        self.dinputs[self.inputs <= 0 ] = 0
    # Calculate predictions for outputs
    def predictions ( self , outputs ):
        return outputs


# Softmax activation
class Activation_Softmax :
    # Forward pass
    def forward(self, inputs, training):

        # Remember input values
        self.inputs = inputs
        # Get unnormalized probabilities
        exp_values = np.exp(inputs - np.max(inputs, axis=1,
                                            keepdims=True))

        # Normalize them for each sample
        probabilities = exp_values / np.sum(exp_values, axis=1,
                                            keepdims=True)
        self.output = probabilities

    # Backward pass
    def backward(self, dvalues):

        # Create uninitialized array
        self.dinputs = np.empty_like(dvalues)
        # Enumerate outputs and gradients
        for index, (single_output, single_dvalues) in \
                enumerate(zip(self.output, dvalues)):
            # Flatten output array
            single_output = single_output.reshape(- 1, 1)
            # Calculate Jacobian matrix of the output and
            jacobian_matrix = np.diagflat(single_output) - \
                              np.dot(single_output, single_output.T)
            # Calculate sample-wise gradient
            # and add it to the array of sample gradients
            self.dinputs[index] = np.dot(jacobian_matrix,
                                         single_dvalues)

    # Calculate predictions for outputs
    def predictions(self, outputs):
        return np.argmax(outputs, axis=1)



# Sigmoid activation
class Activation_Sigmoid :
    # Forward pass
    def forward ( self , inputs , training ):
        # Save input and calculate/save output
        # of the sigmoid function
        self.inputs = inputs
        self.output = 1 / ( 1 + np.exp( - inputs))

    # Backward pass
    def backward ( self , dvalues ):
        # Derivative - calculates from output of the sigmoid function
        self.dinputs = dvalues * ( 1 - self.output) * self.output

    # Calculate predictions for outputs
    def predictions ( self , outputs ):
        return (outputs > 0.5 ) * 1


# Linear activation
class Activation_Linear :
    # Forward pass
    def forward(self, inputs, training):
        # Just remember values
        self.inputs = inputs
        self.output = inputs

    # Backward pass
    def backward(self, dvalues):
        # derivative is 1, 1 * dvalues = dvalues - the chain rule
        self.dinputs = dvalues.copy()

    # Calculate predictions for outputs
    def predictions(self, outputs):
        return outputs



# SGD optimizer
class Optimizer_SGD :
    # Initialize optimizer - set settings,
    # learning rate of 1. is default for this optimizer
    def __init__ ( self , learning_rate = 1. , decay = 0. , momentum = 0. ):
        self.learning_rate = learning_rate
        self.current_learning_rate = learning_rate
        self.decay = decay
        self.iterations = 0
        self.momentum = momentum
    # Call once before any parameter updates
    def pre_update_params ( self ):
        if self.decay:
            self.current_learning_rate = self.learning_rate * ( 1. / ( 1. + self.decay * self.iterations))
        # Update parameters

    def update_params ( self , layer ):
        # If we use momentum
        if self.momentum:
            # If layer does not contain momentum arrays, create them
            # filled with zeros
                if not hasattr (layer, 'weight_momentums' ):
                    layer.weight_momentums = np.zeros_like(layer.weights)
                    # If there is no momentum array for weights
                    # The array doesn't exist for biases yet either.
                    layer.bias_momentums = np.zeros_like(layer.biases)

                # Build weight updates with momentum - take previous
                # updates multiplied by retain factor and update with
                # current gradients
                weight_updates = self.momentum * layer.weight_momentums - self.current_learning_rate * layer.dweights
                layer.weight_momentums = weight_updates
                # Build bias updates
                bias_updates = self.momentum * layer.bias_momentums - self.current_learning_rate * layer.dbiases
                layer.bias_momentums = bias_updates
                # Vanilla SGD updates (as before momentum update)
        else:
                weight_updates = - self.current_learning_rate * layer.dweights
                bias_updates = - self.current_learning_rate * layer.dbiases
        # Update weights and biases using either
        # vanilla or momentum updates
        layer.weights += weight_updates
        layer.biases += bias_updates


    # Call once after any parameter updates
    def post_update_params ( self ):
        self.iterations += 1


# Adagrad optimizer
class Optimizer_Adagrad :
    # Initialize optimizer - set settings
    def __init__ ( self , learning_rate = 1. , decay = 0. , epsilon = 1e-7 ):
        self.learning_rate = learning_rate
        self.current_learning_rate = learning_rate
        self.decay = decay
        self.iterations = 0
        self.epsilon = epsilon
    # Call once before any parameter updates
    def pre_update_params ( self ):
        if self.decay:
            self.current_learning_rate = self.learning_rate *  ( 1. / ( 1. + self.decay * self.iterations))

    # Update parameters
    def update_params ( self , layer ):
        # If layer does not contain cache arrays,
        # create them filled with zeros
        if not hasattr (layer, 'weight_cache' ):
            layer.weight_cache = np.zeros_like(layer.weights)
            layer.bias_cache = np.zeros_like(layer.biases)
        # Update cache with squared current gradients
        layer.weight_cache += layer.dweights ** 2
        layer.bias_cache += layer.dbiases ** 2
        # Vanilla SGD parameter update + normalization
        # with square rooted cache
        layer.weights += - self.current_learning_rate * layer.dweights / (np.sqrt(layer.weight_cache) + self.epsilon)
        layer.biases += - self.current_learning_rate * layer.dbiases / (np.sqrt(layer.bias_cache) + self.epsilon)

    # Call once after any parameter updates
    def post_update_params ( self ):
        self.iterations += 1



# RMSprop optimizer
class Optimizer_RMSprop :
    # Initialize optimizer - set settings
    def __init__ (self, learning_rate=0.001, decay=0., epsilon=1e-7, rho=0.9):
        self.learning_rate = learning_rate
        self.current_learning_rate = learning_rate
        self.decay = decay
        self.iterations = 0
        self.epsilon = epsilon
        self.rho = rho

    # Call once before any parameter updates
    def pre_update_params ( self ):
        if self.decay:
            self.current_learning_rate = self.learning_rate * ( 1. / ( 1. + self.decay * self.iterations))

    # Update parameters
    def update_params(self, layer):
        # If layer does not contain cache arrays,
        # create them filled with zeros
        if not hasattr (layer, 'weight_cache' ):
            layer.weight_cache = np.zeros_like(layer.weights)
            layer.bias_cache = np.zeros_like(layer.biases)
            # Update cache with squared current gradients
        layer.weight_cache = self.rho * layer.weight_cache + (1 - self.rho) * layer.dweights ** 2
        layer.bias_cache = self.rho * layer.bias_cache + (1 - self.rho) * layer.dbiases ** 2
            # Vanilla SGD parameter update + normalization
            # with square rooted cache
        layer.weights += - self.current_learning_rate * layer.dweights / (np.sqrt(layer.weight_cache) + self.epsilon)
        layer.biases += - self.current_learning_rate * layer.dbiases / (np.sqrt(layer.bias_cache) + self.epsilon)

    # Call once after any parameter updates
    def post_update_params ( self ):
        self.iterations += 1

# Adam optimizer
class Optimizer_Adam :
    # Initialize optimizer - set settings
    def __init__ ( self , learning_rate = 0.001 , decay = 0. , epsilon = 1e-7 ,
    beta_1 = 0.9 , beta_2 = 0.999 ):
        self.learning_rate = learning_rate
        self.current_learning_rate = learning_rate
        self.decay = decay
        self.iterations = 0
        self.epsilon = epsilon
        self.beta_1 = beta_1
        self.beta_2 = beta_2
    # Call once before any parameter updates
    def pre_update_params ( self ):
        if self.decay:
            self.current_learning_rate = self.learning_rate * \
            ( 1. / ( 1. + self.decay * self.iterations))

    # Update parameters
    def update_params ( self , layer ):
        # If layer does not contain cache arrays,
        # create them filled with zeros
        if not hasattr (layer, 'weight_cache' ):
            layer.weight_momentums = np.zeros_like(layer.weights)
            layer.weight_cache = np.zeros_like(layer.weights)
            layer.bias_momentums = np.zeros_like(layer.biases)
            layer.bias_cache = np.zeros_like(layer.biases)

        # Update momentum with current gradients
        layer.weight_momentums = self.beta_1 * \
        layer.weight_momentums + \
        ( 1 - self.beta_1) * layer.dweights
        layer.bias_momentums = self.beta_1 * \
        layer.bias_momentums + \
        ( 1 - self.beta_1) * layer.dbiases
        # Get corrected momentum
        # self.iteration is 0 at first pass
        # and we need to start with 1 here
        weight_momentums_corrected = layer.weight_momentums / \
        ( 1 - self.beta_1 ** (self.iterations + 1 ))
        bias_momentums_corrected = layer.bias_momentums / \
        ( 1 - self.beta_1 ** (self.iterations + 1 ))
        # Update cache with squared current gradients
        layer.weight_cache = self.beta_2 * layer.weight_cache + \
        ( 1 - self.beta_2) * layer.dweights ** 2
        layer.bias_cache = self.beta_2 * layer.bias_cache + \
        ( 1 - self.beta_2) * layer.dbiases ** 2
        # Get corrected cache
        weight_cache_corrected = layer.weight_cache / \
        ( 1 - self.beta_2 ** (self.iterations + 1 ))
        bias_cache_corrected = layer.bias_cache / \
        ( 1 - self.beta_2 ** (self.iterations + 1 ))
        # Vanilla SGD parameter update + normalization
        # with square rooted cache
        layer.weights += - self.current_learning_rate * \
        weight_momentums_corrected / \
        (np.sqrt(weight_cache_corrected) +
        self.epsilon)
        layer.biases += - self.current_learning_rate * \
        bias_momentums_corrected / \
        (np.sqrt(bias_cache_corrected) +
        self.epsilon)

    # Call once after any parameter updates
    def post_update_params ( self ):
        self.iterations += 1



# Common loss class
class Loss :
    # Regularization loss calculation
    def regularization_loss ( self ):
        # 0 by default
        regularization_loss = 0
        # Calculate regularization loss
        # iterate all trainable layers
        for layer in self.trainable_layers:
            # L1 regularization - weights
            # calculate only when factor greater than 0
            if layer.weight_regularizer_l1 > 0 :
                regularization_loss += layer.weight_regularizer_l1 * np.sum(np.abs(layer.weights))
            # L2 regularization - weights
            if layer.weight_regularizer_l2 > 0 :
                regularization_loss += layer.weight_regularizer_l2 * np.sum(layer.weights * layer.weights)
            # L1 regularization - biases
            # calculate only when factor greater than 0
            if layer.bias_regularizer_l1 > 0 :
                regularization_loss += layer.bias_regularizer_l1 * np.sum(np.abs(layer.biases))
            # L2 regularization - biases
            if layer.bias_regularizer_l2 > 0 :
                regularization_loss += layer.bias_regularizer_l2 * np.sum(layer.biases * layer.biases)
        return regularization_loss


    # Set/remember trainable layers
    def remember_trainable_layers ( self , trainable_layers ):
        self.trainable_layers = trainable_layers

    # Calculates the data and regularization losses
    # given model output and ground truth values
    def calculate ( self , output , y , * , include_regularization = False ):
        # Calculate sample losses
        sample_losses = self.forward(output, y)
        # Calculate mean loss
        data_loss = np.mean(sample_losses)
        # If just data loss - return it
        if not include_regularization:
            return data_loss
        # Return the data and regularization losses
        return data_loss, self.regularization_loss()

    # Reset variables for accumulated loss
    def new_pass(self):
        self.accumulated_sum = 0
        self.accumulated_count = 0

    # Calculates accumulated loss
    def calculate_accumulated(self, *, include_regularization=False):

        # Calculate mean loss
        data_loss = self.accumulated_sum / max(self.accumulated_count,1)
        # If just data loss - return it
        if not include_regularization:
            return data_loss
        # Return the data and regularization losses
        return data_loss, self.regularization_loss()

# Cross-entropy loss
class Loss_CategoricalCrossentropy ( Loss ):
    # Forward pass
    def forward ( self , y_pred , y_true ):
        # Number of samples in a batch
        samples = len (y_pred)
        # Clip data to prevent division by 0
        # Clip both sides to not drag mean towards any value
        y_pred_clipped = np.clip(y_pred, 1e-7 , 1 - 1e-7 )
        # Probabilities for target values -
        # only if categorical labels
        if len (y_true.shape) == 1 :
            correct_confidences = y_pred_clipped[
            range (samples),
            y_true
            ]

        # Mask values - only for one-hot encoded labels
        elif len (y_true.shape) == 2 :
            correct_confidences = np.sum(
            y_pred_clipped * y_true,
            axis = 1
            )

        # Losses
        negative_log_likelihoods = - np.log(correct_confidences)
        return negative_log_likelihoods

    # Backward pass
    def backward ( self , dvalues , y_true ):
        # Number of samples
        samples = len (dvalues)
        # Number of labels in every sample
        # We'll use the first sample to count them
        labels = len (dvalues[ 0 ])
        # If labels are sparse, turn them into one-hot vector
        if len (y_true.shape) == 1 :
            y_true = np.eye(labels)[y_true]
        # Calculate gradient
        self.dinputs = - y_true / dvalues
        # Normalize gradient
        self.dinputs = self.dinputs / samples
        # Softmax classifier - combined Softmax activation
        # and cross-entropy loss for faster backward step


class Activation_Softmax_Loss_CategoricalCrossentropy ():
    # Backward pass
    def backward ( self , dvalues , y_true ):
        # Number of samples
        samples = len (dvalues)
        # If labels are one-hot encoded,
        # turn them into discrete values
        if len (y_true.shape) == 2 :
            y_true = np.argmax(y_true, axis = 1 )
        # Copy so we can safely modify
        self.dinputs = dvalues.copy()
        # Calculate gradient
        self.dinputs[ range (samples), y_true] -= 1
        # Normalize gradient
        self.dinputs = self.dinputs / samples



# Binary cross-entropy loss
class Loss_BinaryCrossentropy ( Loss ):
    # Forward pass
    def forward ( self , y_pred , y_true ):
        # Clip data to prevent division by 0
        # Clip both sides to not drag mean towards any value
        y_pred_clipped = np.clip(y_pred, 1e-7 , 1 - 1e-7 )
        # Calculate sample-wise loss
        sample_losses = - (y_true * np.log(y_pred_clipped) +
        ( 1 - y_true) * np.log( 1 - y_pred_clipped))
        sample_losses = np.mean(sample_losses, axis =- 1 )
        # Return losses
        return sample_losses
    # Backward pass
    def backward ( self , dvalues , y_true ):
        # Number of samples
        samples = len (dvalues)
        # Number of outputs in every sample
        # We'll use the first sample to count them
        outputs = len (dvalues[ 0 ])
        # Clip data to prevent division by 0
        # Clip both sides to not drag mean towards any value
        clipped_dvalues = np.clip(dvalues, 1e-7 , 1 - 1e-7 )
        # Calculate gradient
        self.dinputs = - (y_true / clipped_dvalues -
        ( 1 - y_true) / ( 1 - clipped_dvalues)) / outputs
        # Normalize gradient
        self.dinputs = self.dinputs / samples




# Mean Squared Error loss
class Loss_MeanSquaredError ( Loss ): # L2 loss
    # Forward pass
    def forward ( self , y_pred , y_true ):
        # Calculate loss
        sample_losses = np.mean((y_true - y_pred) ** 2 , axis =- 1 )
        # Return losses
        return sample_losses

    # Backward pass
    def backward ( self , dvalues , y_true ):
        # Number of samples
        samples = len (dvalues)
        # Number of outputs in every sample
        # We'll use the first sample to count them
        outputs = len (dvalues[ 0 ])
        # Gradient on values
        self.dinputs = - 2 * (y_true - dvalues) / outputs
        # Normalize gradient
        self.dinputs = self.dinputs / samples



# Mean Absolute Error loss
class Loss_MeanAbsoluteError ( Loss ): # L1 loss
    def forward ( self , y_pred , y_true ):
        # Calculate loss
        sample_losses = np.mean(np.abs(y_true - y_pred), axis =- 1 )
        # Return losses
        return sample_losses
    # Backward pass
    def backward ( self , dvalues , y_true ):
        # Number of samples
        samples = len (dvalues)
        # Number of outputs in every sample
        # We'll use the first sample to count them
        outputs = len (dvalues[ 0 ])
        # Calculate gradient
        self.dinputs = np.sign(y_true - dvalues) / outputs
        # Normalize gradient
        self.dinputs = self.dinputs / samples



# Common accuracy class
class Accuracy :
    # Calculates an accuracy
    # given predictions and ground truth values
    def calculate ( self , predictions , y ):
        # Get comparison results
        comparisons = self.compare(predictions, y)
        # Calculate an accuracy
        accuracy = np.mean(comparisons)
        # Return accuracy
        return accuracy

    # Calculates accumulated accuracy
    def calculate_accumulated(self):

        # Calculate an accuracy
        accuracy = self.accumulated_sum / max(self.accumulated_count,1)
        # Return the data and regularization losses
        return accuracy


    def new_pass(self):
        self.accumulated_sum = 0
        self.accumulated_count = 0


# Accuracy calculation for classification model
class Accuracy_Categorical ( Accuracy ):
    # No initialization is needed
    def init ( self , y ):
       pass
    # Compares predictions to the ground truth values
    def compare ( self , predictions , y ):
        if len (y.shape) == 2 :
            y = np.argmax(y, axis = 1 )
        return predictions == y


# Accuracy calculation for regression model
class Accuracy_Regression ( Accuracy ):
    def __init__ ( self ):
        # Create precision property
        self.precision = None
    # Calculates precision value
    # based on passed in ground truth values
    def init ( self , y , reinit = False ):
        if self.precision is None or reinit:
            self.precision = np.std(y) / 250

    # Compares predictions to the ground truth values
    def compare ( self , predictions , y ):
        return np.absolute(predictions - y) < self.precision



# Model class
class Model :
    def __init__ ( self ):
        # Create a list of network objects
        self.layers = []
        # Softmax classifier's output object
        self.softmax_classifier_output = None


    # Add objects to the model
    def add ( self , layer ):
        self.layers.append(layer)


    # Set loss, optimizer and accuracy
    def set ( self , * , loss , optimizer , accuracy ):
        self.loss = loss
        self.optimizer = optimizer
        self.accuracy = accuracy


    # Finalize the model
    def finalize ( self ):
        # Create and set the input layer
        self.input_layer = Layer_Input()
        # Count all the objects
        layer_count = len (self.layers)
        # Initialize a list containing trainable layers:
        self.trainable_layers = []
        # Iterate the objects
        for i in range (layer_count):
            # If it's the first layer,
            # the previous layer object is the input layer
            if i == 0 :
                self.layers[i].prev = self.input_layer
                self.layers[i].next = self.layers[i + 1 ]
            # All layers except for the first and the last
            elif i < layer_count - 1 :
                self.layers[i].prev = self.layers[i - 1 ]
                self.layers[i].next = self.layers[i + 1 ]

            # The last layer - the next object is the loss
            # Also let's save aside the reference to the last object
            # whose output is the model's output
            else :
                self.layers[i].prev = self.layers[i - 1 ]
                self.layers[i].next = self.loss
                self.output_layer_activation = self.layers[i]
            # If layer contains an attribute called "weights",
            # it's a trainable layer -
            # add it to the list of trainable layers
            # We don't need to check for biases -
            # checking for weights is enough
            if hasattr (self.layers[i], 'weights' ):
                self.trainable_layers.append(self.layers[i])
                # Update loss object with trainable layers
                self.loss.remember_trainable_layers(
                self.trainable_layers
                )
        # If output activation is Softmax and
        # loss function is Categorical Cross-Entropy
        # create an object of combined activation
        # and loss function containing
        # faster gradient calculation
        if isinstance (self.layers[ - 1 ], Activation_Softmax) and \
        isinstance (self.loss, Loss_CategoricalCrossentropy):
            # Create an object of combined activation
            # and loss functions
            self.softmax_classifier_output = \
            Activation_Softmax_Loss_CategoricalCrossentropy()


    # Train the model

    def train ( self , X , y , * , epochs = 1 , batch_size = None , print_every = 1 ,
        validation_data = None ):
        # Initialize accuracy object
        self.accuracy.init(y)
        # Default value if batch size is not being set
        train_steps = 1
        # If there is validation data passed,
        # set default number of steps for validation as well
        if validation_data is not None:
            validation_steps = 1
        # For better readability
        X_val, y_val = validation_data
        # Calculate number of steps
        if batch_size is not None:
            train_steps = len(X) // batch_size
        # Dividing rounds down. If there are some remaining
        # data but not a full batch, this won't include it
        # Add `1` to include this not full batch
        if train_steps * batch_size < len(X):
            train_steps += 1
        if validation_data is not None:
            validation_steps = len(X_val) // batch_size

        # Dividing rounds down. If there are some remaining
        # data but nor full batch, this won't include it
        # Add `1` to include this not full batch
        if validation_steps * batch_size < len(X_val):
            validation_steps += 1
        # Main training loop
        for epoch in range(1, epochs + 1):
            # Print epoch number
            print(
            'epoch' + str(epoch))
            # Reset accumulated values in loss and accuracy objects
            self.loss.new_pass()
            self.accuracy.new_pass()

            # Iterate over steps
            for step in range(train_steps):
                # If batch size is not set -
                # train using one step and full dataset
                if batch_size is None:
                    batch_X = X
                    batch_y = y
                    # Otherwise slice a batch
                else:
                    batch_X = X[step * batch_size:(step + 1) * batch_size]
                    batch_y = y[step * batch_size:(step + 1) * batch_size]
                # Perform the forward pass
                output = self.forward(batch_X, training=True)
                # Calculate loss
                data_loss, regularization_loss = self.loss.calculate(output, batch_y, include_regularization=True)
                loss = data_loss + regularization_loss
                # Get predictions and calculate an accuracy
                predictions = self.output_layer_activation.predictions(output)
                accuracy = self.accuracy.calculate(predictions, batch_y)
                # Perform backward pass
                self.backward(output, batch_y)

                # Optimize (update parameters)
                self.optimizer.pre_update_params()
                for layer in self.trainable_layers:
                    self.optimizer.update_params(layer)
                self.optimizer.post_update_params()
                # Print a summary
                if not step % print_every or step == train_steps - 1:
                    print(
                    'step:' +str(step) +
                    'acc:' +str(accuracy)  +
                    'loss:'  + str(loss) +
                    'data_loss:' + str(data_loss) +
                    'reg_loss:' + str(regularization_loss) +
                    'lr:' + str(self.optimizer.current_learning_rate))

                # Get and print epoch loss and accuracy
            epoch_data_loss, epoch_regularization_loss = self.loss.calculate_accumulated(include_regularization=True)
            epoch_loss = epoch_data_loss + epoch_regularization_loss
            epoch_accuracy = self.accuracy.calculate_accumulated()
            print('training, ' +
                'acc:' + str(epoch_accuracy)+
                'loss:' + str(epoch_loss)+
                'data_loss:' + str(epoch_data_loss) +
                'reg_loss:' + str(epoch_regularization_loss) +
                'lr:' +  str(self.optimizer.current_learning_rate))
                # If there is the validation data
            if validation_data is not None:
                # Reset accumulated values in loss
                # and accuracy objects
                self.loss.new_pass()
                self.accuracy.new_pass()
                # Iterate over steps
                for step in range(validation_steps):
                    # If batch size is not set -
                    # train using one step and full dataset
                    if batch_size is None:
                        batch_X = X_val
                        batch_y = y_val

                    # Otherwise slice a batch
                    else:
                        batch_X = X_val[
                                  step * batch_size:(step + 1) * batch_size
                                  ]
                        batch_y = y_val[
                                  step * batch_size:(step + 1) * batch_size
                                  ]
                # Perform the forward pass
                output = self.forward(batch_X, training=False)
                # Calculate the loss
                self.loss.calculate(output, batch_y)
                # Get predictions and calculate an accuracy
                predictions = self.output_layer_activation.predictions(output)
                self.accuracy.calculate(predictions, batch_y)


            # Get and print validation loss and accuracy
            validation_loss = self.loss.calculate_accumulated()
            validation_accuracy = self.accuracy.calculate_accumulated()
            # Print a summary
            print(
            'validation, ' +
            'acc:' + str(validation_accuracy)+
            'loss:' + str(validation_loss) )


    # Performs forward pass
    def forward( self , X , training ):
        # Call forward method on the input layer
        # this will set the output property that
        # the first layer in "prev" object is expecting
        self.input_layer.forward(X, training)
        # Call forward method of every object in a chain
        # Pass output of the previous object as a parameter
        for layer in self.layers:
            layer.forward(layer.prev.output, training)
        # "layer" is now the last object from the list,
        # return its output
        return layer.output


    # Performs backward pass
    def backward ( self , output , y ):
    # If softmax classifier
        if self.softmax_classifier_output is not None :
            # First call backward method
            # on the combined activation/loss
            # this will set dinputs property
            self.softmax_classifier_output.backward(output, y)
            # Since we'll not call backward method of the last layer
            # which is Softmax activation
            # as we used combined activation/loss
            # object, let's set dinputs in this object
            self.layers[ - 1 ].dinputs = \
            self.softmax_classifier_output.dinputs
            # Call backward method going through
            # all the objects but last
            # in reversed order passing dinputs as a parameter
            for layer in reversed (self.layers[: - 1 ]):
                layer.backward(layer.next.dinputs)
            return

        # First call backward method on the loss
        # this will set dinputs property that the last
        # layer will try to access shortly
        self.loss.backward(output, y)
        # Call backward method going through all the objects
        # in reversed order passing dinputs as a parameter
        for layer in reversed (self.layers):
            layer.backward(layer.next.dinputs)

    # Retrieves and returns parameters of trainable layers
    def get_parameters(self):

        # Create a list for parameters
        parameters = []
        # Iterable trainable layers and get their parameters
        for layer in self.trainable_layers:
            parameters.append(layer.get_parameters())
        # Return a list
        return parameters

    # Updates the model with new parameters
    def set_parameters(self, parameters):

        # Iterate over the parameters and layers
        # and update each layers with each set of the parameters
        for parameter_set, layer in zip(parameters,
                                        self.trainable_layers):
            layer.set_parameters(*parameter_set)

    # Saves the parameters to a file
    def save_parameters(self, path):

        # Open a file in the binary-write mode
        # and save parameters into it
        with open(path, 'wb') as f:
            pickle.dump(self.get_parameters(), f)

    # Loads the weights and updates a model instance with them
    def load_parameters(self, path):

        # Open file in the binary-read mode,
        # load weights and update trainable layers
        with open(path, 'rb') as f:
            self.set_parameters(pickle.load(f))

    # Saves the model
    def save(self, path):

        # Make a deep copy of current model instance
        model = copy.deepcopy(self)
        # Reset accumulated values in loss and accuracy objects
        model.loss.new_pass()
        model.accuracy.new_pass()
        # Remove data from the input layer
        # and gradients from the loss object
        model.input_layer.__dict__.pop('output', None)
        model.loss.__dict__.pop('dinputs', None)
        # For each layer remove inputs, output and dinputs properties
        for layer in model.layers:
            for property in ['inputs', 'output', 'dinputs',
                             'dweights', 'dbiases']:
                layer.__dict__.pop(property, None)
        # Open a file in the binary-write mode and save the model
        with open(path, 'wb') as f:
            pickle.dump(model, f)

    # Loads and returns a model
    @staticmethod
    def load(path):

        # Open file in the binary-read mode, load a model
        with open(path, 'rb') as f:
            model = pickle.load(f)
        # Return a model
        return model

    def predict(self, X, *, batch_size=None):

        # Default value if batch size is not being set
        prediction_steps = 1
        # Calculate number of steps
        if batch_size is not None:
            prediction_steps = len(X) // batch_size

        # Dividing rounds down. If there are some remaining
        # data but not a full batch, this won't include it
        # Add `1` to include this not full batch
            if prediction_steps * batch_size < len(X):
                prediction_steps += 1
        # Model outputs
        output = []
        # Iterate over steps
        for step in range(prediction_steps):
        # If batch size is not set -
            # train using one step and full dataset
            if batch_size is None:
                batch_X = X
            # Otherwise slice a batch
            else:
                batch_X = X[step * batch_size:(step + 1) * batch_size]
            # Perform the forward pass
            batch_output = self.forward(batch_X, training=False)
            # Append batch prediction to the list of predictions
            output.append(batch_output)
        # Stack and return results
        return np.vstack(output)


"""
# Create dataset
X, y = spiral_data( samples = 1000 , classes = 3 )
X_test, y_test = spiral_data( samples = 100 , classes = 3 )
# Instantiate the model
model = Model()
# Add layers
model.add(Layer_Dense( 2 , 512 , weight_regularizer_l2 = 5e-4 , bias_regularizer_l2 = 5e-4 ))
model.add(Activation_ReLU())
model.add(Layer_Dropout( 0.1 ))
model.add(Layer_Dense( 512 , 3 ))
model.add(Activation_Softmax())
# Set loss, optimizer and accuracy objects
model.set(
loss = Loss_CategoricalCrossentropy(),
optimizer = Optimizer_Adam( learning_rate = 0.05 , decay = 5e-5 ),
accuracy = Accuracy_Categorical()
)
# Finalize the model
model.finalize()
# Train the model
model.train(X, y, validation_data = (X_test, y_test),
epochs = 10000 , print_every = 100 )

"""


def load_mnist_dataset ( dataset , path ):
    # Scan all the directories and create a list of labels
    labels = os.listdir(os.path.join(path, dataset))
    # Create lists for samples and labels
    X = []
    y = []

    # For each label folder
    for label in labels:
        # And for each image in given folder
        for file in os.listdir(os.path.join(path, dataset, label)):
            # Read the image
            image = cv2.imread(
            os.path.join(path, dataset, label, file),
            cv2.IMREAD_UNCHANGED)
            # And append it and a label to the lists
            X.append(image)
            y.append(label)
    # Convert the data to proper numpy arrays and return
    return np.array(X), np.array(y).astype( 'uint8' )












# MNIST dataset (train + test)
def create_data_mnist ( path ):
    # Load both sets separately
    X, y = load_mnist_dataset( 'train' , path)
    X_test, y_test = load_mnist_dataset( 'test' , path)
    # And return all the data
    return X, y, X_test, y_test


"""

# Create dataset
X, y, X_test, y_test = create_data_mnist( 'fashion_mnist_images' )
# Shuffle the training dataset
keys = np.array( range (X.shape[ 0 ]))
np.random.shuffle(keys)
X = X[keys]
y = y[keys]
# Scale and reshape samples
X = (X.reshape(X.shape[ 0 ], - 1 ).astype(np.float32) - 127.5 ) / 127.5
X_test = (X_test.reshape(X_test.shape[ 0 ], - 1 ).astype(np.float32) -
127.5 ) / 127.5
# Instantiate the model
model = Model()

# Add layers
model.add(Layer_Dense(X.shape[ 1 ], 128 ))
model.add(Activation_ReLU())
model.add(Layer_Dense( 128 , 128 ))
model.add(Activation_ReLU())
model.add(Layer_Dense( 128 , 10 ))
model.add(Activation_Softmax())
# Set loss, optimizer and accuracy objects
model.set(
loss = Loss_CategoricalCrossentropy(),
optimizer = Optimizer_Adam( decay = 1e-4 ),
accuracy = Accuracy_Categorical()
)
# Finalize the model
model.finalize()
# Train the model
model.train(X, y, validation_data = (X, y),
epochs = 15 , batch_size = 128 , print_every = 100 )
model.save( 'fashion_mnist.model' )




"""

# Label index to label name relation
fashion_mnist_labels = {
0 : 'T-shirt/top' ,
1 : 'Trouser' ,
2 : 'Pullover' ,
3 : 'Dress' ,
4 : 'Coat' ,
5 : 'Sandal' ,
6 : 'Shirt' ,
7 : 'Sneaker' ,
8 : 'Bag' ,
9 : 'Ankle boot'
}
# Read an image
image_data = cv2.imread( 'pants.png' , cv2.IMREAD_GRAYSCALE)
# Resize to the same size as Fashion MNIST images
image_data = cv2.resize(image_data, ( 28 , 28 ))

# Invert image colors
image_data = 255 - image_data
# Reshape and scale pixel data
image_data = (image_data.reshape( 1 , - 1 ).astype(np.float32) -
127.5 ) / 127.5
# Load the model
model = Model.load( 'fashion_mnist.model' )
# Predict on the image
confidences = model.predict(image_data)
# Get prediction instead of confidence levels
predictions = model.output_layer_activation.predictions(confidences)
# Get label name from label index
prediction = fashion_mnist_labels[predictions[ 0 ]]
print (prediction)
temp = []

for i in range(len(image_data)):
    temp.append(image_data[i])
    if i%28==0:
        print(temp)
        temp=[]











"""

Implementation of ML introductory algorithms visulisation as ancillary
Author - mr frozen (or just frozen)
Github - github.com/mrfrozen97
Mail - mrfrozenpeak@gmail.com

This file has class that performs linear regression algorithm to the input data and can also visualise it .
The packages used here are numpy(for efficient calculations) and matplotlib(for data visualisation)



Classes:
Linear Regression- It carries out regression model
plot model- It carries out data visualisation

Equation of best fit line y = bx + a
In linear regression we want to minimise the squared error which is ( y(best_fit) - y(original) )^2

y(o) -> y coordinate of original line
x(o) -> x coordinate of original line
y(b) -> y coordinate of best fit line
x(b) -> x coordinate of best fit line
x(mean) -> mean of all x(o) values
y(mean) -> mean of all y(o) values







#Mathematics.........................................................................................................................

Let F = ∑ (y(b) - y(o))^2
F = ∑ (y(0) - y(b))^2

We also know that y(b) = b*x(b) + a
So, F = ∑ (y(o) - b*x(b) - a)^2

Minimising error.....
a and b are our unkowns and so to find them we partially differentiate them w.r.t. F

∂F/∂a = 0     -(1)
∂F/∂b = 0     -(2)

Let y(b) - y(o) = u

∂F/∂a = ∂F/∂u * ∂u/∂a -(3)
∂F/∂b = ∂F/∂u * ∂u/∂b -(4)

F = u^2
∂F/∂u = 2u

u = ∑y(o) - ∑(bx(b) - a)
∂u/∂a = -1

From  3 and 1,
0 = -2*u
0 = -2*(∑y(o) -∑(bx(b) - a))
na = (∑y(o) -∑(bx(b))
a = (y(mean) - b*x(mean)) -(5)

Now,
∂u/∂a = -∑x(b)

From 4 and 1,
0 = -2*u*∑x(b)
0 = (∑y(o)∑x(b)  -∑(bx(b) - a)*∑x(b) )
∑x(b)*na = ∑y(o)∑x(b)  -∑bx(b)
From 5

∑x(b)*n(y(mean) - bx(mean)) = ∑y(o)∑x(b)  -∑bx(b)
Solving this


b =  ∑(y(o)x(a))(mean) -  (y(mean)*x(mean)) /  ∑x(o)*x(mean) - x(mean)**2                        ###############.....IMP
a = y(mean) - b*x(mean)                                                                          ###############.....IMP

squared mean error is defined as
SME = 1 - ( ∑(y(o)-y(mean))^2 / ∑(y(o)-y(b))^2 )                                                 ###############.....IMP

#End of math. Maybe, Never skip maths................................................................







Attributes:
............................
____________________________
b                  -> indicates slope of best fit line
a                  -> indicates x intercept of the  best fit line
x                  -> input x coordinates
y                  -> input y coordinates
x_mean             -> mean of x values
y_mean             -> mean of y values
squared_mean_error -> squared error to check accuracy
showGrid           -> Boolean which indicates weather grid should be shown or not
sizeX              -> Horizontal size of graph in inches
sizeY              -> Vertical size of graph in inches
labelx_count       -> number of labels on x axis
labely_count       -> number of labels on y axis
title_color        -> color of title of plot
label_default      -> Boolean which tells if the label is default or changed by user
xlabel_color       -> color of x label
ylabel_color       -> color of y label
title              -> title text
x_label            -> x label text
y_label            -> y label text
xlabel_size        -> x_label font size of text
ylabel_size        -> y_label font size of text
title_size         -> title font size of text
model              -> this is the model obeject i.e. linear regression object in case we need any parent class variables
unknown_points_size-> this specifies the size of unknown points if any on the graph
unknown_points_color-> this specifies the color of unknown points if any on the graph
unknown_points_apl]ha->  this specifies the transparency of unknown points if any on the graph


#####.....IMP.......####
color_dict -> Dictionary that contains wide range of colors in the form of dictionary. The keys are the names of colors
              , values are the hex codes of that color. You can get this dict using get_color_dict function()
You can print this dict i.e. print(object.get_color_dict()) to know which colors are available.



You are most welcomed to improve this code:
You can pull request for this code at

github.com/mrfrozen97/               (In spider algorithms repo)
or
Email - mrfrozenpeak@gmail.com

"""





import numpy as np
import matplotlib.pyplot as plt
import math
from frozenSpider import spiderAlgorithmResources as res







"""

Class that performs linear regression
It takes in the x and y coordinates of the data
It then calculates the value of b and a using numpy array functions
It returns values of a and b
It also returns values of squared mean error in squared_error function

"""

class Linear_regression:


    def __init__(self):
        self.a = 0.0
        self.b = 0.0
        self.squared_mean_error = 0.0
        self.x_mean = 0
        self.y_mean = 0
        self.size = 1
        self.x = []
        self.y = []
        self.calculated_y = []
        self.calculated_x = []


    #Function to calcute and return a, b values....................................

    def bestFit(self, x_coordinates, y_coordinates):

        size_coordinates = np.size(x_coordinates)
        x_coordinates_mean = np.sum(x_coordinates)/size_coordinates
        y_coordinates_mean = np.sum(y_coordinates)/size_coordinates

        self.x = x_coordinates
        self.y = y_coordinates
        self.size = size_coordinates
        self.x_mean = x_coordinates_mean
        self.y_mean = y_coordinates_mean

        bestFint_numerator = (np.sum(np.multiply(x_coordinates, y_coordinates))/size_coordinates) - (x_coordinates_mean*y_coordinates_mean)
        bestFint_denominator = (np.sum(np.multiply(x_coordinates,x_coordinates))/size_coordinates) - (x_coordinates_mean**2)
        self.b = bestFint_numerator/bestFint_denominator
        self.a = y_coordinates_mean - (self.b*x_coordinates_mean)
        return self.a, self.b




    def bestFit_logrithemic(self, x_coordinates, y_coordinates, base=math.e):

        y_coordinates = [math.log(x, base) for x in x_coordinates]

        size_coordinates = np.size(x_coordinates)
        x_coordinates_mean = np.sum(x_coordinates)/size_coordinates
        y_coordinates_mean = np.sum(y_coordinates)/size_coordinates

        self.x = x_coordinates
        self.y = y_coordinates
        self.size = size_coordinates
        self.x_mean = x_coordinates_mean
        self.y_mean = y_coordinates_mean

        bestFint_numerator = (np.sum(np.multiply(x_coordinates, y_coordinates))/size_coordinates) - (x_coordinates_mean*y_coordinates_mean)
        bestFint_denominator = (np.sum(np.multiply(x_coordinates,x_coordinates))/size_coordinates) - (x_coordinates_mean**2)
        self.b = bestFint_numerator/bestFint_denominator
        self.a = y_coordinates_mean - (self.b*x_coordinates_mean)
        return self.a, self.b








    #function that calculates and returns squarred error which can be attributed to the accuracy og our best fit line............

    def squared_error(self):

        mean_y_array = [self.y_mean for x in range(self.size)]
        bestFit_y = [(self.b*x + self.a) for x in self.x]

       # print(mean_y_array)
       # print(bestFit_y)

        squared_error_denominator = np.sum(np.square(np.subtract(bestFit_y, mean_y_array)))
        squared_error_numerator = np.sum(np.square(np.subtract(bestFit_y, self.y)))


        self.squared_mean_error = 1 - (squared_error_numerator/squared_error_denominator)

        return self.squared_mean_error

    def find_unknowns(self, x):
        x = list(x)
        self.calculated_x = x
        for i in x:
            self.calculated_y.append(self.b * i + self.a)
        return self.calculated_y













"""

This is child class of the linear regression model class
It basically plots the result obtained by performing linear regression vs the original data
The parameters such as title, x/y labels , their colors, their size, plot colors, x/y label values, grid, etc can be set
There are function to set these values as well as to set the size of graph, to save it, etc.
All the default values of the above variables are already set so you can directly plot entire graph with just one line of code......

"""

class plot_model(Linear_regression):


    #All the default value are already set but they can all be reset.........................

    def __init__(self, model):
        self.showGrid = True
        self.sizeX = 10
        self.labelx_count = 10
        self.labely_count = 10
        self.title_color = "#FF0000"
        self.xlabel_color = "#663399"
        self.ylabel_color = "#663399"
        self.label_default = True
        self.sizeY = 6
        self.title = ""
        self.x_label = ""
        self.y_label = ""
        self.xlabel_size = 15
        self.ylabel_size = 15
        self.title_size = 18
        self.model = model
        self.color_dict = res.Resources.get_color_dict()
        self.unknown_points_size = 20
        self.unknown_points_color = "Orange"
        self.unknown_points_alpha = 0.8





    #Function to get dictionary to access the different options avaliable for the colors.........................

    def get_color_dict(self):
        return self.color_dict






    #function to set vlues for title, x/y label and values of labels in case.........................
    def set_marker_properties(self,unknown_points_alpha=0.8, unknown_points_color = "Orange", unknown_points_size = 20, xlabel_count=17, label_default=True, ylabel_count=12, title_size=15, xlabel_size=10, ylabel_size=10, title="Linear Regression", x_label="x coordinates", y_label="y coordinates", title_color="#FF0000", xlabel_color="#663399", ylabel_color="#663399"):
        self.title_size = title_size
        self.xlabel_size = xlabel_size
        self.ylabel_size = ylabel_size
        self.x_label = x_label
        self.y_label = y_label
        self.title = title
        self.xlabel_color = xlabel_color
        self.ylabel_color = ylabel_color
        self.title_color = title_color
        self.labelx_count =xlabel_count
        self.labely_count =ylabel_count
        self.label_default = label_default
        self.unknown_points_size = unknown_points_size
        self.unknown_points_color = unknown_points_color
        self.unknown_points_alpha = unknown_points_alpha









   #Sets the display size which will be equal to size of image if it is being svaed.........................

    def set_display_size(self, sizeX, sizeY):
        self.sizeX = sizeX
        self.sizeY = sizeY






    #function to create x,y values according to the best fit line equation................................
    def create_best_fit_line(self):
        y_coords_line = []
        x_coords_line = list(self.model.x)
        for i in self.model.x:
            y_coords_line.append(self.model.b*i + self.model.a)
        #print(self.model.calculated_x)
        for i in self.model.calculated_x:
            x_coords_line.append(i)
            y_coords_line.append(self.model.b*i + self.model.a)
        #print(len(self.model.x) - len(self.model.y))
        return x_coords_line, y_coords_line







    #This function creates f\default ticks for x and y axis ....
    def generate_ticks(self, label_x, label_y):
        if self.label_default or len(label_x)==0 or len(label_y)==0:
             min_y = min(self.model.y)
             min_x = min(self.model.x)
             max_y = max(max(self.model.y), max(self.model.calculated_y))
             max_y += int(max_y*0.3)
             max_x = max(max(self.model.x), max(self.model.calculated_x))
             max_x += int(max_x * 0.2)
            # dif_y = max(1, int(len(self.model.y)/self.labely_count))
             dif_y = max(1, int((max_y-min_y) / self.labely_count))
             #dif_x = max(1, int(len(self.model.x)/self.labelx_count))
             dif_x = max(1, int((max_x-min_x) / self.labelx_count))
             print(dif_x, dif_y)
             x_labels_default = []
             y_labels_default = []


             while min_x<=max_x:
                 x_labels_default.append(min_x)
                 min_x+=dif_x
             while min_y<max_y:

                 y_labels_default.append(min_y)
                 min_y += dif_y


            # print(x_labels_default)
            # print(y_labels_default)

             plt.xticks(x_labels_default)
             plt.yticks(y_labels_default)

        else:
            plt.xticks(label_x)
            plt.yticks(label_y)









    #main fuction thats plots the graph and saved it if path provided.........................

    def plot_model(self,label="Points",alpha=0.6, point_size=25,label_x=[], label_y=[],line_color='Purple',unknown_points_label="Unknown points",line_label='Best Fit line', point_color='DeepSkyBlue', save_fig_path='dont'):


        #These are the coordinates to plot the line which is inclusion of both known and calculated points
        x_bestFitLine_coords, y_bestFitLine_coords = self.create_best_fit_line()

        plt.plot(x_bestFitLine_coords, y_bestFitLine_coords, color=self.color_dict[line_color], label=line_label)                     #plot the best fit line
        plt.scatter(self.model.x, self.model.y, zorder=3, label=label, s=point_size, color=self.color_dict[point_color], alpha=alpha)       #plot all the points
        plt.scatter(self.model.calculated_x, self.model.calculated_y, color=self.color_dict[self.unknown_points_color], label=unknown_points_label, zorder = 4, alpha=self.unknown_points_alpha, s=self.unknown_points_size)

        plt.title(self.title,
                  fontdict=res.Resources.title_dict,
                  color=self.title_color)
        plt.xlabel(self.x_label,
                   fontdict=res.Resources.label_dict,
                   color=self.xlabel_color)
        plt.ylabel(self.y_label,
                   fontdict=res.Resources.label_dict,
                   color=self.ylabel_color)

        self.generate_ticks(label_x, label_y)



        plt.legend(loc = 'upper left')

        if self.showGrid:
            plt.grid(color='#cfd8dc', zorder=0)

        figure = plt.gcf()
        figure.set_size_inches(self.sizeX, self.sizeY)


        if not(save_fig_path=='dont') :

              plt.savefig("./" + save_fig_path)#, bbox_inches='tight')
        plt.show()








#Sample code to work with this file of the library
"""

x = [-22, 17, 22, 28, 35]
y = [-9.3, 5, 8, 12, 14]
x1 = [13, 47]


le1 = linear_regression()
a,b =le1.bestFit(x, y)
le1.find_unknowns(x1)


print("Slope = " + str(b))
print("x intercept = " + str(a))
print("Squared error = "+ str(le1.squared_error()))


plot2 = plot_model(le1)
plot2.set_marker_properties(unknown_points_size=80, unknown_points_alpha=0.6, unknown_points_color='Red')
plot2.plot_model(save_fig_path="plot2")


"""


"""

This file has resources class which has all the resources such as hex code for all the colors. It also has some abstract
methods which assign default values to the color/size/e.t.c. to the properties of the graph


#Attributes
color dict -> Dictionary that stores name and hex values of the colors
label_dict -> Default dict with plot label properties
title_dict -> Default dic with plot title properties



#Methods
get_color_dict -> Returns the color dict whenever required


"""


class Resources:

    color_dict = {
        'aliceblue': '#F0F8FF',
        'antiquewhite': '#FAEBD7',
        'aqua': '#00FFFF',
        'aquamarine': '#7FFFD4',
        'azure': '#F0FFFF',
        'beige': '#F5F5DC',
        'bisque': '#FFE4C4',
        'black': '#000000',
        'blanchedalmond': '#FFEBCD',
        'blue': '#0000FF',
        'blueviolet': '#8A2BE2',
        'brown': '#A52A2A',
        'burlywood': '#DEB887',
        'cadetblue': '#5F9EA0',
        'chartreuse': '#7FFF00',
        'chocolate': '#D2691E',
        'coral': '#FF7F50',
        'cornflowerblue': '#6495ED',
        'cornsilk': '#FFF8DC',
        'crimson': '#DC143C',
        'cyan': '#00FFFF',
        'darkblue': '#00008B',
        'darkcyan': '#008B8B',
        'darkgoldenrod': '#B8860B',
        'darkgray': '#A9A9A9',
        'darkgreen': '#006400',
        'darkgrey': '#A9A9A9',
        'darkkhaki': '#BDB76B',
        'darkmagenta': '#8B008B',
        'darkolivegreen': '#556B2F',
        'darkorange': '#FF8C00',
        'darkorchid': '#9932CC',
        'darkred': '#8B0000',
        'darksalmon': '#E9967A',
        'darkseagreen': '#8FBC8F',
        'darkslateblue': '#483D8B',
        'darkslategray': '#2F4F4F',
        'darkslategrey': '#2F4F4F',
        'darkturquoise': '#00CED1',
        'darkviolet': '#9400D3',
        'deeppink': '#FF1493',
        'deepskyblue': '#00BFFF',
        'dimgray': '#696969',
        'dimgrey': '#696969',
        'dodgerblue': '#1E90FF',
        'firebrick': '#B22222',
        'floralwhite': '#FFFAF0',
        'forestgreen': '#228B22',
        'fuchsia': '#FF00FF',
        'gainsboro': '#DCDCDC',
        'ghostwhite': '#F8F8FF',
        'gold': '#FFD700',
        'goldenrod': '#DAA520',
        'gray': '#808080',
        'grey': '#808080',
        'green': '#008000',
        'greenyellow': '#ADFF2F',
        'honeydew': '#F0FFF0',
        'hotpink': '#FF69B4',
        'indianred': '#CD5C5C',
        'indigo': '#4B0082',
        'ivory': '#FFFFF0', 'khaki': '#F0E68C',
        'lavender': '#E6E6FA',
        'lavenderblush': '#FFF0F5',
        'lawngreen': '#7CFC00',
        'lemonchiffon': '#FFFACD',
        'lightblue': '#ADD8E6',
        'lightcoral': '#F08080',
        'lightcyan': '#E0FFFF',
        'lightgoldenrodyellow': '#FAFAD2',
        'lightgray': '#D3D3D3',
        'lightgreen': '#90EE90',
        'lightgrey': '#D3D3D3',
        'lightpink': '#FFB6C1',
        'lightsalmon': '#FFA07A',
        'lightseagreen': '#20B2AA',
        'lightskyblue': '#87CEFA',
        'lightslategray': '#778899',
        'lightslategrey': '#778899',
        'lightsteelblue': '#B0C4DE',
        'lightyellow': '#FFFFE0',
        'lime': '#00FF00',
        'limegreen': '#32CD32',
        'linen': '#FAF0E6',
        'magenta': '#FF00FF',
        'maroon': '#800000',
        'mediumaquamarine': '#66CDAA',
        'mediumblue': '#0000CD',
        'mediumorchid': '#BA55D3',
        'mediumpurple': '#9370DB',
        'mediumseagreen': '#3CB371',
        'mediumslateblue': '#7B68EE',
        'mediumspringgreen': '#00FA9A',
        'mediumturquoise': '#48D1CC',
        'mediumvioletred': '#C71585',
        'midnightblue': '#191970',
        'mintcream': '#F5FFFA',
        'mistyrose': '#FFE4E1',
        'moccasin': '#FFE4B5',
        'navajowhite': '#FFDEAD',
        'navy': '#000080',
        'oldlace': '#FDF5E6',
        'olive': '#808000',
        'olivedrab': '#6B8E23',
        'orange': '#FFA500',
        'orangered': '#FF4500',
        'orchid': '#DA70D6',
        'palegoldenrod': '#EEE8AA',
        'palegreen': '#98FB98',
        'paleturquoise': '#AFEEEE',
        'palevioletred': '#DB7093',
        'papayawhip': '#FFEFD5',
        'peachpuff': '#FFDAB9',
        'peru': '#CD853F',
        'pink': '#FFC0CB',
        'plum': '#DDA0DD',
        'powderblue': '#B0E0E6',
        'purple': '#800080',
        'red': '#FF0000',
        'rosybrown': '#BC8F8F',
        'royalblue': '#4169E1',
         'saddlebrown': '#8B4513',
        'salmon': '#FA8072',
        'sandybrown': '#F4A460',
        'seagreen': '#2E8B57',
        'seashell': '#2E8B57',
        'sienna': '#A0522D',
        'silver': '#C0C0C0',
        'skyblue': '#87CEEB',
        'slateblue': '#6A5ACD',
        'slategray': '#708090',
        'slategrey': '#708090',
        'snow': '#FFFAFA',
        'springgreen': '#00FF7F',
        'steelblue': '#4682B4',
        'tan': '#D2B48C',
        'teal': '#008080',
        'thistle': '#D8BFD8',
        'tomato': '#FF6347',
        'turquoise': '#40E0D0',
        'violet': '#EE82EE',
        'wheat': '#F5DEB3',
        'white': '#FFFFFF',
        'whitesmoke': '#F5F5F5',
        'yellow': '#FFFF00',
        'yellowgreen': '#9ACD32'

    }

    label_dict = {"fontsize": 15}

    title_dict = {"fontsize": 20}


    @staticmethod
    def get_color_dict():

        return Resources.color_dict


#print(Resources.get_color_dict())




#Example code to use the linear regression of the library

"""

m = 0.43
c = -5.34
x = [i for i in range(1000)]
y = []
for i in x:
    y.append(m*i + c + (- 0.5 + random.random())*(min(int(i/3), 200)))

x1 = [104, 205, 625]

le1 = sp.linear_regression()
a,b =le1.bestFit(x, y)
le1.find_unknowns(x1)


print("Slope = " + str(b))
print("x intercept = " + str(a))
print("Squared error = "+ str(le1.squared_error()))


plot2 = sp.plot_model(le1)
plot2.set_marker_properties(unknown_points_size=80, unknown_points_alpha=0.6, unknown_points_color='Red')
plot2.plot_model(save_fig_path="plot2")


"""




#LOR TSNE
"""


model = TSNE(n_components=2)
tsne_data = model.fit_transform(x_train)

#tsne_data = np.array(tsne_data).T
class1 = []
class2 = []


for i in range(len(y_train)):
    if y_train[i] == 0:
        class1.append(tsne_data[i])
    else:
        class2.append(tsne_data[i])

class1 = np.array(class1).T
class2 = np.array(class2).T

plt.scatter(class1[0], class1[1], color="#FFA500")
plt.scatter(class2[0], class2[1], color="#800080")
plt.show()





tsne_data = model.fit_transform(x_test)

#tsne_data = np.array(tsne_data).T
class1 = []
class2 = []


for i in range(len(lor.y_calculated)):
    if lor.y_calculated[i] == 0:
        class1.append(x_test[i])
    else:
        class2.append(x_test[i])

class1 = np.array(class1).T
class2 = np.array(class2).T

plt.scatter(class1[0], class1[1], color="#FFA500", alpha=0.6)
plt.scatter(class2[0], class2[1], color="#800080", alpha=0.6)
plt.show()




model1 = TSNE(n_components=3)
tsne_data = model1.fit_transform(x_train)

#tsne_data = np.array(tsne_data).T
class1 = []
class2 = []


for i in range(len(y_train)):
    if y_train[i] == 0:
        class1.append(tsne_data[i])
    else:
        class2.append(tsne_data[i])

class1 = np.array(class1).T
class2 = np.array(class2).T




fig = plt.figure()
ax = plt.axes(projection='3d')
ax.set_facecolor((0, 0, 0))
ax.scatter3D(class1[0], class1[1], class1[2], color="#800080", alpha=0.9)
ax.scatter3D(class2[0], class2[1], class2[2], color="#FFA500", alpha=0.9)
plt.show()




"""



















"""


bc1 = datasets.load_wine()
x1, y1 = bc1.data, bc1.target
print(y1)
x_train1, x_test1, y_train1, y_test1 = train_test_split(x1, y1, test_size=0.2, random_state=1234)


lor1 = lore.LogisticRegression(learning_rate=0.01, n_iters=10000)
lor1.best_fit(x_train1, y_train1)
prediction1 = lor1.predict(x_test1)

accuracy1 = np.sum(prediction1 == y_test1)/ len(prediction1)

print(accuracy1)


plotly1 = ninja_technique()
plotly11 = ninja_technique()
plotly1.calculate(lor1, x_train1, y_train1 )
plotly1.plot()
plotly11.calculate(lor1, x_test1, y_test1)
plotly11.plot()


datav = Logistic_regression_plot2D(lor1)
datav.plot_2D_visuals(plot_test_data=False)



bc1 = datasets.load_iris()
x1, y1 = bc1.data, bc1.target

x_train1, x_test1, y_train1, y_test1 = train_test_split(x1, y1, test_size=0.2, random_state=1234)


lor1 = lore.LogisticRegression(learning_rate=0.01, n_iters=100000)
lor1.best_fit(x_train1, y_train1)
prediction1 = lor1.predict(x_test1)

accuracy1 = np.sum(prediction1 == y_test1)/ len(prediction1)

print(accuracy1)


plotly1 = ninja_technique()
plotly11 = ninja_technique()
plotly1.calculate(lor1, x_train1, y_train1)
plotly1.plot()
plotly11.calculate(lor1, x_test1, y_test1)
plotly11.plot()



"""


from frozenSpider import spiderAlgorithmLoR as lor
from frozenSpider import Data_visualisation as dv
from sklearn import datasets as ds
from sklearn.model_selection import train_test_split
import numpy as np

dataset = ds.load_breast_cancer()

x, y = dataset.data, dataset.target

x_train, x_test, y_train, y_test = train_test_split(x, y, test_size=0.2, random_state=1234)


model = lor.LogisticRegression(learning_rate=0.01)
model.best_fit(x_train, y_train)

prediction = model.predict(x_test)

accuracy = np.sum(prediction == y_test)/ len(prediction)

print(accuracy)

plot = lor.plot_model(model)
plot.set_marker_properties()




#All the below properties are set default. So you need to change them, only when you want to change it...................................
plot.set_marker_properties(
                              label_default=True,
                              calculated_point_class1_label = "Calculated class 1",
                              calculated_point_class0_label="Calculated class 0",
                              calculated_point_alpha = 0.5,
                              calculated_point_class1_color='red',
                              calculated_point_class0_color="cyan",
                              title_size=15, xlabel_size=10,
                              ylabel_size=10, title="Output vs dimention",
                              y_label="y coordinates",
                              title_color="#FF0000",
                              xlabel_color="#663399",
                              ylabel_color="#663399",
                              )







#ALl the parameters are already defined with default values so not required to reset if not required.................
plot.plot_model(
                display_graph=False,
                class1_color='purple',
                class0_color='orange',
                alpha=0.6,
                save_fig_file='dont',
                display_calculated_points = True,
                #save_fig_file=""                                       # Pass string of the path
                )



datav = dv.Logistic_regression_plot(model)
datav.plot_3D_visuals(plot_test_data=True)

datav = dv.Logistic_regression_plot2D(model)
datav.plot_2D_visuals(plot_test_data=True)








# Funtion to check if matrix is symmetric
def check_symmetrc(arr, n):
    for i in range(n):
        for j in range(i, n):
            if arr[i][j] != arr[j][i]:
                return False  # false if elements do not match

    return True  # True if elements match


def add_matrix(a, b):
    sum_mat = []
    if len(a) != len(b) and len(a[0]!=len(b[0])):
        print("Matrix cannot be added")

    else:
        for i in range(len(a)):
            temp_sum = []
            for j in range(len(a[i])):
                temp_sum.append(a[i][j] + b[i][j])
            sum_mat.append(temp_sum)

    return sum_mat


def transpose(l1):
    l2 = []
    # iterate over list l1 to the length of an item
    for i in range(len(l1[0])):
        # print(i)
        row = []
        for item in l1:
            row.append(item[i])
        l2.append(row)
    return l2


def multiply(a, b):
    b = transpose(b)
    if len(a) != len(b[0]):
        print("Matrix cannot be multiplied")

    else:
        mul_ans = []
        for i in range(len(a)):
            temp = []
            for j in range(len(b[0])):
                temp.append(sum([a[i][x]*b[j][x] for x in range(len(b[0]))]))
            mul_ans.append(temp)


    return mul_ans

# Number of rows and columns input
n = int(input("Enter number of rows: "))
m = int(input("Enter numbe of columns: "))

# Matrix elements input
print("Enter matrix elements")
arr = []  # list to store the matrix
for i in range(n):
    print("Enter row " + str(i) + ": ", end=" ")
    temp = [int(i) for i in input().split()]
    if len(temp) != m:  # Check for invalid number of inputs
        print("Invalid Input")
        n = m + 1
        break
    arr.append(temp)

# Condition for matix to be symmetric
if n == m:
    if check_symmetrc(arr, n):  # Check for matching elements
        print("It is a symmetric matrix")
    else:
        print("Matrix is not symmetric")
else:
    print("Matrix is not symmetric")




n = int(input("Enter number of rows: "))
m = int(input("Enter numbe of columns: "))

# Matrix elements input
print("Enter matrix elements")
arr1 = []  # list to store the matrix
for i in range(n):
    print("Enter row " + str(i) + ": ", end=" ")
    temp = [int(i) for i in input().split()]
    if len(temp) != m:  # Check for invalid number of inputs
        print("Invalid Input")
        n = m + 1
        break
    arr1.append(temp)


print("Sum is \n" + str(add_matrix(arr, arr1)))
print("Multiply is \n" + str(multiply(arr, arr1)))


"""
This File implements q_learning algorithm i.e. deep_q_learning

Deep_Q_Learning ha three parts
1)Game
2)Agent
3)Model

This File takes care of Agent and Model
So u can build any game meeting some basic requirements to implement deep_q_learning


#Game Class should have 3 methods to work with this file
1)Reset
-> This method should reset the game and store any info if required

2)game_step
-> This method takes in action and move the game forward and should return the reward, game_over(boolean) and score

3)get_state
-> This method returns current state of the game which is given to neural nets as an input


Then create a object of Deep_q_learning and pass game class object
then call train method to train the model


At the end of the code is an example of using this class


"""



import pygame
import os
import random
from collections import deque
from enum import Enum
import torch
import numpy as np
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as f
from collections import namedtuple
import matplotlib.pyplot as plt





class Deep_q_learning:

    def __init__(self, game,
                 max_memory=50_000, epsilon=0, gamma=0.9, lr=0.001,
                 outputs=3, inputs=11, hidden_layers=[256],
                 batch_size=100,  average_last=1,
                 show_graph=False, plot_every=10, plot_display_time=2,
                 plot_save_every=None, plot_save_path="", save_model_path="Deep_Q_Model1",
                 score_threshold=None, load_model_path=None):

        self.plot_scores = []
        self.plot_scoresx = []
        self.plot_mean_scores = []
        self.model_save_path = save_model_path
        self.total_score = 0
        self.score_threshold = score_threshold
        self.record = 0
        self.outputs=outputs
        self.plot_save_path = plot_save_path
        self.plot_save_every = plot_save_every
        self.average_last = average_last
        self.show_graph = show_graph
        self.agent = Agent(max_memory=max_memory, gamma=gamma, epsilon=epsilon, lr=lr,
                           outputs=outputs, inputs=inputs, hidden_layers=hidden_layers,
                           batch_size=batch_size, load_model_path=load_model_path)
        self.game = game
        self.game.agent = self.agent
        self.valid_game = True
        self.plot_display_time=plot_display_time
        self.plot_every = plot_every
        self.left_score_ptr = 0
        if not(callable(getattr(self.game, 'reset', None))):
            self.valid_game = False
            print("Warning: Game object class does not contain method reset\n" +
                  "reset method should reset/restart the game\n")
        else:
            pass

        if not(callable(getattr(self.game, 'play_step', None))):
            self.valid_game = False
            print("Warning: Game object class does not contain method play_step\n" +
                  "play_step method takes in action and returns following values\n" +
                  "1)Reward\n2)game_over (boolean)\n3)score\n")
        else:
            pass

        if not(callable(getattr(self.game, 'get_state', None))):
            self.valid_game = False
            print("Warning: Game object class does not contain get_state\n" +
                  "play_step method should return an array of states that are given input to the neural nets\n")
        else:
            pass





    def plot_graph(self, n_iters):


        if (n_iters%self.plot_every == 0) or (n_iters%self.plot_save_every==0) and n_iters!=0:
            #print('haello!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!', n_iters%self.plot_save_every, self.plot_every)
            plt.style.use("dark_background")
            plt.ion()
            plt.plot(self.plot_scoresx, self.plot_scores, label='Score')
            plt.plot(self.plot_scoresx, self.plot_mean_scores, label= 'Avg Score')
            plt.ylabel("Score", fontdict={'size':10, 'color':'blue'})
            plt.xlabel("Number of Games", fontdict={'size':10, 'color':'blue'})
            plt.legend()
            if n_iters % self.plot_every == 0:
                plt.show()
            if self.plot_save_every and n_iters%self.plot_save_every==0:
                plt.savefig(self.plot_save_path+str(int(n_iters/self.plot_save_every)))
            plt.pause(self.plot_display_time)
            plt.close()




    def play_game(self, end_game_score=None):
        while True:
            final_move = [0]*self.outputs
            state = self.game.get_state()
            state0 = torch.tensor(state, dtype=torch.float)
            prediction = self.agent.model(state0)
            move = torch.argmax(prediction).item()
            final_move[move] = 1
            reward, game_over, score = game.play_step(final_move)
            if game_over:
                game.reset()
            if end_game_score:
                if score>=end_game_score:
                    print("Score: " + str(score))
                    break



    def train_model(self):
        if self.valid_game:
            n_iters = 0
            while True:
                # get old state
                state_old = self.game.get_state()

                # get move
                final_move = self.agent.get_action(state_old)

                # perform move and get new state
                reward, done, score = self.game.play_step(final_move)
                state_new = self.game.get_state()  # /8

                self.agent.train_short_memory(state_old, final_move, reward, state_new, done)
                self.agent.remember(state_old, final_move, reward, state_new, done)


                if done or (self.score_threshold and score>self.score_threshold):
                    n_iters+=1
                    # train long memory
                    self.game.reset()
                    self.agent.n_games += 1
                    self.agent.train_long_memory()
                    self.score = score

                    if self.score > self.record:
                        self.record = self.score
                        self.agent.model.save(filename=self.model_save_path)
                    print('Game', self.agent.n_games, 'Score', self.score, 'Record:', self.record)
                    self.plot_scores.append(score)
                    self.plot_scoresx.append(n_iters)
                    self.total_score += score
                    if n_iters>self.average_last:
                        self.total_score-=self.plot_scores[self.left_score_ptr]
                        self.left_score_ptr += 1
                    self.plot_mean_scores.append(self.total_score / min(n_iters, self.average_last))


                    self.plot_graph(n_iters)
        else:
            print("The game object class does not contain the required methods\nCheck if class contains reset, train_step, methods ")









class Linear_QNet(nn.Module):

    def __init__(self, input_size=10, hidden_sizes=[256],
                 hidden_activation='relu', output_size=3,
                 dropout=None, input_activation='relu',
                 output_activation='linear'):
        super().__init__()
        temp_size = hidden_sizes[0]
        self.layers = nn.ModuleList()
        self.layers.append(nn.Linear(input_size, temp_size))
        if len(hidden_sizes) > 1:
            for i in range(1, len(hidden_sizes)):
                self.layers.append(nn.Linear(temp_size, i))
                temp_size = i

        self.layers.append(nn.Linear(temp_size, output_size))

    def forward(self, x):
        for i in self.layers[:-1]:
            x = f.relu(i(x))

        x = self.layers[-1](x)
        return x

    def save(self, filename='model.pth'):
        model_folder_path = './model'
        if not os.path.exists(model_folder_path):
            os.makedirs(model_folder_path)

        filename = os.path.join(model_folder_path, filename)
        torch.save(self.state_dict(), filename)

    def load_model(self, filepath="model.pth"):
        self.load_state_dict(torch.load(filepath))
        self.eval()


class QTrainer:
        def __init__(self, model, lr, gamma):
            self.lr = lr
            self.gamma = gamma
            self.model = model
            self.criterion = nn.MSELoss()
            self.optimizer = optim.Adam(model.parameters(), lr=self.lr)
            self.MINIBATCH_SIZE = 100

        def train_step(self, state, action, reward, next_state, done):

            state = torch.tensor(state, dtype=torch.float)
            action = torch.tensor(action, dtype=torch.long)
            next_state = torch.tensor(next_state, dtype=torch.float)
            reward = torch.tensor(reward, dtype=torch.float)

            if len(state.shape) == 1:
                state = torch.unsqueeze(state, 0)
                next_state = torch.unsqueeze(next_state, 0)
                action = torch.unsqueeze(action, 0)
                reward = torch.unsqueeze(reward, 0)
                done = (done,)

            pred = self.model(state)

            target = pred.clone()
            for i in range(len(done)):
                Q_new = reward[i]
                if not done[i]:
                    Q_new = reward[i] + self.gamma * torch.max(self.model(next_state[i]))

                target[i][torch.argmax(action).item()] = Q_new

            self.optimizer.zero_grad()
            loss = self.criterion(target, pred)
            loss.backward()

            self.optimizer.step()

class Agent:

        def __init__(self, max_memory=50_000, epsilon=0, gamma=0.9, lr=0.001,
                     outputs=3, inputs=11, hidden_layers=[256],
                     batch_size=100, load_model_path=None):
            self.n_games = 0
            self.epsilon = epsilon  # Randomness to agent
            self.memory = deque(maxlen=max_memory)
            self.batch_size = batch_size
            self.gamma = gamma
            self.outputs=outputs

            # model

            self.model = Linear_QNet(inputs, hidden_layers, output_size=outputs)
            if load_model_path != None:
                self.model.load_model("mode/pixel1_deep_q_model")
            self.trainer = QTrainer(self.model, lr=lr, gamma=self.gamma)

        def remember(self, state, action, reward, next_state, done):
            self.memory.append((state, action, reward, next_state, done))  # popleft if MAX_MEMORY is reached

        def train_long_memory(self):

            if len(self.memory) > self.batch_size:
                mini_sample = random.sample(self.memory, self.batch_size)  # list of tuples
            else:
                mini_sample = self.memory

            states, actions, rewards, next_states, dones = zip(*mini_sample)
            self.trainer.train_step(states, actions, rewards, next_states, dones)

        def train_short_memory(self, state, action, reward, next_state, done):
            self.trainer.train_step(state, action, reward, next_state, done)

        def get_action(self, state):
            # random moves: tradeoff exploration / exploitation
            self.epsilon = 80 - self.n_games
            final_move = [0]*self.outputs
            if random.randint(0, 200) < self.epsilon:
                move = random.randint(0, 2)
                final_move[move] = 1
            else:
                state0 = torch.tensor(state, dtype=torch.float)
                prediction = self.model(state0)
                move = torch.argmax(prediction).item()
                final_move[move] = 1

            return final_move

















########################################################################################################################




#Example of using this file is shown below




"""

pygame.init()
font = pygame.font.Font(pygame.font.get_default_font(), 25)


# font = pygame.font.SysFont('arial', 25)

class Direction(Enum):
    RIGHT = 1
    LEFT = 2
    UP = 3
    DOWN = 4


Point = namedtuple('Point', 'x, y')

# rgb colors
WHITE = (255, 255, 255)
RED = (200, 0, 0)
BLUE1 = (0, 0, 255)
BLUE2 = (0, 100, 255)
BLACK = (0, 0, 0)

BLOCK_SIZE = 20
SPEED = 200


class SnakeGameAI:

    def __init__(self, w=640, h=480):
        self.w = w
        self.h = h
        # init display
        self.display = pygame.display.set_mode((self.w, self.h))
        pygame.display.set_caption('Snake')
        self.clock = pygame.time.Clock()
        self.agent = None
        self.hscore = 0
        self.score = 0
        self.reset()

    def reset(self):
        # init game state
        self.direction = Direction.RIGHT

        self.head = Point(self.w / 2, self.h / 2)
        self.snake = [self.head,
                      Point(self.head.x - BLOCK_SIZE, self.head.y),
                      Point(self.head.x - (2 * BLOCK_SIZE), self.head.y)]


        if self.score > self.hscore:
            self.hscore = self.score

        self.score = 0
        self.food = None
        self._place_food()
        self.frame_iteration = 0




    def get_state(self):
        head = self.snake[0]
        point_l = Point(head.x - 20, head.y)
        point_r = Point(head.x + 20, head.y)
        point_u = Point(head.x, head.y - 20)
        point_d = Point(head.x, head.y + 20)

        dir_l = self.direction == Direction.LEFT
        dir_r = self.direction == Direction.RIGHT
        dir_u = self.direction == Direction.UP
        dir_d = self.direction == Direction.DOWN

        state = [
            # Danger straight
            (dir_r and self.is_collision(point_r)) or
            (dir_l and self.is_collision(point_l)) or
            (dir_u and self.is_collision(point_u)) or
            (dir_d and self.is_collision(point_d)),

            # Danger right
            (dir_u and self.is_collision(point_r)) or
            (dir_d and self.is_collision(point_l)) or
            (dir_l and self.is_collision(point_u)) or
            (dir_r and self.is_collision(point_d)),

            # Danger left
            (dir_d and self.is_collision(point_r)) or
            (dir_u and self.is_collision(point_l)) or
            (dir_r and self.is_collision(point_u)) or
            (dir_l and self.is_collision(point_d)),

            # Move direction
            dir_l,
            dir_r,
            dir_u,
            dir_d,

            # Food location
            self.food.x < self.head.x,  # food left
            self.food.x > self.head.x,  # food right
            self.food.y < self.head.y,  # food up
            self.food.y > self.head.y  # food down
        ]

        return np.array(state, dtype=int)




    def _place_food(self):
        x = random.randint(0, (self.w - BLOCK_SIZE) // BLOCK_SIZE) * BLOCK_SIZE
        y = random.randint(0, (self.h - BLOCK_SIZE) // BLOCK_SIZE) * BLOCK_SIZE
        self.food = Point(x, y)
        if self.food in self.snake:
            self._place_food()

    def play_step(self, action):
        self.frame_iteration += 1
        # 1. collect user input
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                pygame.quit()
                quit()

        # 2. move
        self._move(action)  # update the head
        self.snake.insert(0, self.head)

        # 3. check if game over
        reward = 0
        game_over = False
        if self.is_collision() or self.frame_iteration > 100 * len(self.snake):
            game_over = True
            reward = -10
            return reward, game_over, self.score

        # 4. place new food or just move
        if self.head == self.food:
            self.score += 1
            reward = 10
            self._place_food()
        else:
            self.snake.pop()
            #reward-=0.05
        # 5. update ui and clock
        self._update_ui()
        self.clock.tick(SPEED)
        # 6. return game over and score
        return reward, game_over, self.score

    def is_collision(self, pt=None):
        if pt is None:
            pt = self.head
        # hits boundary
        if pt.x > self.w - BLOCK_SIZE or pt.x < 0 or pt.y > self.h - BLOCK_SIZE or pt.y < 0:
            return True
        # hits itself
        if pt in self.snake[1:]:
            return True

        return False

    def _update_ui(self):
        self.display.fill(BLACK)

        for pt in self.snake:
            pygame.draw.rect(self.display, BLUE1, pygame.Rect(int(pt.x), int(pt.y), BLOCK_SIZE, BLOCK_SIZE))
            pygame.draw.rect(self.display, BLUE2, pygame.Rect(int(pt.x + 4), int(pt.y) + 4, 12, 12))

        pygame.draw.rect(self.display, RED, pygame.Rect(self.food.x, self.food.y, BLOCK_SIZE, BLOCK_SIZE))

        text = font.render("Score: " + str(self.score), True, WHITE)
        self.display.blit(text, [0, 0])
        text = font.render("Games: " + str(self.agent.n_games), True, WHITE)
        self.display.blit(text, [140, 0])
        text = font.render("Highest Score: " + str(self.hscore), True, WHITE)
        self.display.blit(text, [310, 0])
        pygame.display.flip()

    def _move(self, action):
        # [straight, right, left]

        clock_wise = [Direction.RIGHT, Direction.DOWN, Direction.LEFT, Direction.UP]
        idx = clock_wise.index(self.direction)

        if np.array_equal(action, [1, 0, 0]):
            new_dir = clock_wise[idx]  # no change
        elif np.array_equal(action, [0, 1, 0]):
            next_idx = (idx + 1) % 4
            new_dir = clock_wise[next_idx]  # right turn r -> d -> l -> u
        else:  # [0, 0, 1]
            next_idx = (idx - 1) % 4
            new_dir = clock_wise[next_idx]  # left turn r -> u -> l -> d

        self.direction = new_dir

        x = self.head.x
        y = self.head.y
        if self.direction == Direction.RIGHT:
            x += BLOCK_SIZE
        elif self.direction == Direction.LEFT:
            x -= BLOCK_SIZE
        elif self.direction == Direction.DOWN:
            y += BLOCK_SIZE
        elif self.direction == Direction.UP:
            y -= BLOCK_SIZE

        self.head = Point(x, y)


def train():
    #agent = Agent()
    game = SnakeGameAI()
    ql_model = Deep_q_learning(game=game, plot_save_every=50, plot_every=50)
    ql_model.train_model()


if __name__=='__main__':
    train()

"""




























import numpy as np

dvalues = np.array([[1., 1., 1.], [2., 2., 2.], [3., 3., 3.]])

weights = np.array([[0.2, 0.8, - 0.5, 1], [0.5, - 0.91, 0.26, - 0.5], [- 0.26, - 0.27, 0.17, 0.87]]).T

# sum weights of given input
# and multiply by the passed in gradient for this neuron


dinputs = np.dot(dvalues[0], weights.T)
print (dinputs)







import numpy as np
from nnfs.datasets import spiral_data
from nnfs.datasets import vertical_data
import matplotlib.pyplot as plt

np.random.seed(0)

class Layer:

    def __init__(self, weights_count=0, node_count=0):

        self.weights_count = weights_count
        self.nodes_count = node_count
        self.weights = np.random.rand(self.weights_count, self.nodes_count)
        self.biases = np.zeros((1, self.nodes_count))
        self.output = []

        #print(self.weights)
        #print(self.biases)

    def forward(self, input):

        self.output = np.dot(input, self.weights) + self.biases


        return self.output


    def relu(self, input):
        #print(list(input))
        return np.maximum(0, input)

    def softMax(self, input):
        ext_values = np.exp(input - np.max(input, axis=1, keepdims=True))
        normalized_exp = ext_values/np.sum(ext_values, axis=1, keepdims=True)
        return normalized_exp





class Loss:

    #def calculate(self):



    def calculate_loss_crossEntropy(self, y_predict, y_target):

        cliped_y_predict = np.clip(y_predict, 1e-7, 1 - 1e-7)

        if len(y_target.shape)==1:
            output = cliped_y_predict[range(len(cliped_y_predict)), y_target]
        elif len(y_target.shape)==2:
            output = np.sum(cliped_y_predict * y_target, axis=1)



        neg_log_output = -np.log(output)
        return np.mean(neg_log_output)






def backward ( self , dvalues ):
    # Create uninitialized array
    self.dinputs = np.empty_like(dvalues)
    # Enumerate outputs and gradients
    for index, (single_output, single_dvalues) in \
            enumerate ( zip (self.output, dvalues)):
        # Flatten output array
        single_output = single_output.reshape( - 1 , 1 )
        # Calculate Jacobian matrix of the output and
        jacobian_matrix = np.diagflat(single_output) - \
        np.dot(single_output, single_output.T)
        # Calculate sample-wise gradient
        # and add it to the array of sample gradients
        self.dinputs[index] = np.dot(jacobian_matrix,
        single_dvalues)








x, y = spiral_data(100, 3)


layer1 = Layer(2, 3)
o1 = layer1.forward(x)
o1 = layer1.relu(o1)
#print(o1)
l2 = Layer(3, 3)
ab = l2.forward(o1)
ab = l2.softMax(ab)
#print(ab)

input = np.array([[1, 2, 3], [-2, -1, 0]])
target = np.array([[1, 0, 0], [1, 1, 0]])
softmax_output = layer1.softMax(ab)

loss = Loss()
lc = loss.calculate_loss_crossEntropy(softmax_output, y)
print(lc)


X, y = vertical_data( samples = 100 , classes = 3 )
plt.scatter(X[:, 0 ], X[:, 1 ], c = y, s = 40 , cmap = 'brg' )
plt.show()














"""
import nnfs
import numpy as np

nnfs.init()


def create_data(points, classes):

    x = np.zeros((points*classes, 2))
    y = np.zeros(points*classes, dtype='uint8')
    for class_number in range(classes):
        ix = range(points*class_number, points*(class_number+1))
        r = np.linspace(0.0, 1, points)
        t = np.linspace(class_number*4, (class_number+1)*4, points) + np.random.randn(points)*0.2
        x[ix] = np.c_[r*np.sin(t*2.5), r*np.cos(t*2.5)]
        y[ix] = class_number

    return x, y
"""




#x, y = create_data(100, 3)





import numpy as np
import nnfs
from nnfs.datasets import spiral_data
nnfs.init()
# Dense layer
class Layer_Dense :
# Layer initialization
    def __init__ ( self , n_inputs , n_neurons ):
    # Initialize weights and biases
        self.weights = 0.01 * np.random.randn(n_inputs, n_neurons)
        self.biases = np.zeros(( 1 , n_neurons))
    # Forward pass
    def forward ( self , inputs ):
    # Calculate output values from inputs, weights and biases
        self.output = np.dot(inputs, self.weights) + self.biases
    # ReLU activation
class Activation_ReLU :
# Forward pass
    def forward ( self , inputs ):
    # Calculate output values from inputs
        self.output = np.maximum( 0 , inputs)

    # Softmax activation
class Activation_Softmax :
# Forward pass
    def forward ( self , inputs ):
    # Get unnormalized probabilities
        exp_values = np.exp(inputs - np.max(inputs, axis = 1 ,
        keepdims = True ))
        # Normalize them for each sample
        probabilities = exp_values / np.sum(exp_values, axis = 1 ,
        keepdims = True )
        self.output = probabilities
    # Common loss class
class Loss :
# Calculates the data and regularization losses
# given model output and ground truth values
    def calculate ( self , output , y ):
    # Calculate sample losses
        sample_losses = self.forward(output, y)
        # Calculate mean loss
        data_loss = np.mean(sample_losses)
    # Return loss
        return data_loss
# Cross-entropy loss
class Loss_CategoricalCrossentropy ( Loss ):
# Forward pass
    def forward ( self , y_pred , y_true ):
    # Number of samples in a batch
        samples = len (y_pred)
        # Clip data to prevent division by 0
        # Clip both sides to not drag mean towards any value
        y_pred_clipped = np.clip(y_pred, 1e-7 , 1 - 1e-7 )

        # Probabilities for target values -
        # only if categorical labels
        if len (y_true.shape) == 1 :
            correct_confidences = y_pred_clipped[
            range (samples),
            y_true
            ]
        # Mask values - only for one-hot encoded labels
        elif len (y_true.shape) == 2 :
            correct_confidences = np.sum(
            y_pred_clipped * y_true,
            axis = 1
            )
        # Losses
        negative_log_likelihoods = - np.log(correct_confidences)
        return negative_log_likelihoods
# Create dataset
X, y = spiral_data( samples = 100 , classes = 3 )
# Create Dense layer with 2 input features and 3 output values
dense1 = Layer_Dense( 2 , 3 )
# Create ReLU activation (to be used with Dense layer):
activation1 = Activation_ReLU()
# Create second Dense layer with 3 input features (as we take output
# of previous layer here) and 3 output values
dense2 = Layer_Dense( 3 , 3 )
# Create Softmax activation (to be used with Dense layer):
activation2 = Activation_Softmax()
# Create loss function
loss_function = Loss_CategoricalCrossentropy()
# Perform a forward pass of our training data through this layer
dense1.forward(X)
# Perform a forward pass through activation function
# it takes the output of first dense layer here
activation1.forward(dense1.output)

# Perform a forward pass through second Dense layer
# it takes outputs of activation function of first layer as inputs
dense2.forward(activation1.output)
# Perform a forward pass through activation function
# it takes the output of second dense layer here
activation2.forward(dense2.output)
# Let's see output of the first few samples:
print (activation2.output[: 5 ])
# Perform a forward pass through loss function
# it takes the output of second dense layer here and returns loss
loss = loss_function.calculate(activation2.output, y)
print(y)
# Print loss value
print ( 'loss:' , loss)




"""
Implementation of ML introductory algorithms visulisation as ancillary
Author - mr frozen (or just frozen)
Github - github.com/mrfrozen97
Mail - mrfrozenpeak@gmail.com



While handling real world data, the major part of the problem arises when we have to visulize the data because the data
has many parameters which we treat as dimensions. It is very difficult to visualise such data and so instead of wrapping
our mind to perplexing multi dimensional space, we use some algorithms which make our lives easier by transforming this
data to a lower dimensional data which can be plotted and easily understood. Surprisingly, this algorithm called as T-sne
is very powerful. Its efficacy can be observed by the accuracy at which it differentiates between the data.




This file has class that can plot 2d as well as 3d model of a object of logarithmic regression class.
The packages used here are matplotlib(for data visualisation)

This is a preliminary file that performs data visualisation. This is one of the first versions. So a lot
will be improved in the future.


Classes:
Logistic Regression plot - It plots 3D T-sne graph of regression model
Logistic Regression plot 2D - It plots 2D T-sne graph of regression model




#Mathematics...........................................................................................................

For time being, I fully do not understand the math behind T-sne so I would just give a overview of the math working
behind T-sne.

It uses distance between 2 points in nth dimentional space that determines how close two points are in lower dimentinal
space.

It uses the normal function to get a exponential decent, so that the far off points are less significant anyways.

It creates a random arrangement of points in lower dimensional space and then by using the closeness factoe of original
distances decides which points should be moved where.


#end math for now.......................................................






Attributes:
............................
____________________________

model = model
train_class0_color -> color of train data class 0 points
train_class1_color -> color of train data class 1 points
test_class0_color  -> color of test data class 0 points
test_class1_color  -> color of test data class 1 points
title              -> set the title of the plot
background_color   -> set the background color of the plot. Default id black
train_class0_label -> label of train data class 0
train_class1_label -> label of train data class 1
test_class0_label  -> label of test data class 0
test_class1_label  -> label of test data class 1
test_alpha         -> alpha/transparency of the test data
train_alpha        -> alpha/transparency of the train data





#####.....IMP.......####
color_dict -> Dictionary that contains wide range of colors in the form of dictionary. The keys are the names of colors
              , values are the hex codes of that color. You can get this dict using get_color_dict function()
You can print this dict i.e. print(object.get_color_dict()) to know which colors are available.





You are most welcomed to improve this code:
You can pull request for this code at

github.com/mrfrozen97/                      (In spiderAlgorithms repo)
or
Email - mrfrozenpeak@gmail.com



"""



import numpy as np
import matplotlib.pyplot as plt
from sklearn.manifold import TSNE
import random
from frozenSpider import spiderAlgorithmResources as res
#import sklearn
import math
#from sklearn import datasets
#from sklearn.model_selection import  train_test_split
#from frozenSpider import spiderAlgorithmLoR as lore
#from frozenSpider import Data_visualisation as dv






#Class that automatically extracts data from logistical regression model and plots 3D visualisation of the points......

class Logistic_regression_plot():



    #Init method to set default values..................................................

    def __init__(self, model):
        self.model = model
        self.train_class0_color = "#800080"
        self.train_class1_color = "#FFA500"
        self.test_class0_color = "#00BFFF"
        self.test_class1_color = "#FF1493"
        self.title = "Logistic regression visulisation"
        self.background_color = (0, 0, 0)
        self.train_class0_label = "class0 train"
        self.train_class1_label = "class1 train"
        self.test_class0_label = "class0 test"
        self.test_class1_label = "class1 test"
        self.test_alpha = 0.6
        self.train_alpha = 0.9
        self.color_dict = res.Resources.get_color_dict()






    #Method to set the properties of the graph................................
    #Not necessary, because all the default values are already set......................................

    def set_plot_properties(self, train_alpha = 0.9,test_alpha = 0.6,test_class1_label = "class1 test",test_class0_label = "class0 test",train_class1_label = "class1 train", train_class0_label = "class0 train", background_color = (0, 0, 0),title = "Logistic regression visulisation",train_class0_color = "#800080", train_class1_color = "#FFA500", test_class0_color = "#00BFFF", test_class1_color = "#FF1493"):
        self.train_class0_color = train_class0_color
        self.train_class1_color = train_class1_color
        self.test_class0_color = test_class0_color
        self.test_class1_color = test_class1_color
        self.title = title
        self.background_color = background_color
        self.train_class0_label = train_class0_label
        self.train_class1_label = train_class1_label
        self.test_class0_label = test_class0_label
        self.test_class1_label = test_class1_label
        self.test_alpha = test_alpha
        self.train_alpha = train_alpha





    #method that does all the calculatons before plotting the graph.....................................................

    def plot_3D_calculations(self, model1, x, y):
        tsne_data = model1.fit_transform(x)
        class1 = []
        class2 = []

        for i in range(len(y)):
            if y[i] == 0:
                class1.append(tsne_data[i])
            else:
                class2.append(tsne_data[i])

        class1 = np.array(class1).T
        class2 = np.array(class2).T
        return class1, class2







    # Method to plot the 2d graph of the calculated values after apllying T-sne.......................................

    def plot_3D_visuals(self, plot_train_data =True, plot_test_data = False, save_fig_path="dont"):
        model1 = TSNE(n_components=3)

        if plot_train_data:
             class1, class2 = self.plot_3D_calculations(model1, self.model.x, self.model.y)
        if plot_test_data:
             class3, class4 = self.plot_3D_calculations(model1, self.model.x_calculated, self.model.y_calculated)

        plt.style.use('dark_background')

        fig = plt.figure(figsize=(12, 12))
        ax = plt.axes(projection='3d')
        ax.text2D(0.30, 0.98, self.title, transform=ax.transAxes)

        ax.set_facecolor(self.background_color)
        if plot_train_data:
            ax.text2D(0.90, 0.98, self.train_class0_label, transform=ax.transAxes, color=self.train_class0_color)
            ax.text2D(0.90, 0.94, self.train_class1_label, transform=ax.transAxes, color=self.train_class1_color)
            ax.scatter3D(class1[0], class1[1], class1[2], color=self.train_class0_color, alpha=self.train_alpha, label = self.train_class0_label)
            ax.scatter3D(class2[0], class2[1], class2[2], color=self.train_class1_color, alpha=self.train_alpha, label= self.train_class1_label)
        if plot_test_data:
            ax.text2D(0.90, 0.90, self.test_class0_label, transform=ax.transAxes, color=self.test_class0_color)
            ax.text2D(0.90, 0.86, self.test_class1_label, transform=ax.transAxes, color=self.test_class1_color)
            ax.scatter3D(class3[0], class3[1], class3[2], color=self.test_class0_color, alpha=self.test_alpha, label=self.test_class0_label)
            ax.scatter3D(class4[0], class4[1], class4[2], color=self.test_class1_color, alpha=self.test_alpha, label=self.test_class1_label)
        if not(save_fig_path=="dont"):
            plt.savefig(save_fig_path + "/3D" + self.title)

        plt.show()
















#Class that automatically extracts data from logistical regression model and plots 2D visualisation of the points......


class Logistic_regression_plot2D():

    # Init method to set default values..................................................

    def __init__(self, model):
        self.model = model
        self.train_class0_color = "#800080"
        self.train_class1_color = "#FFA500"
        self.test_class0_color = "#00BFFF"
        self.test_class1_color = "#FF1493"
        self.title = "Logistic regression visulisation"
        self.background_color = (0, 0, 0)
        self.train_class0_label = "class0 train"
        self.train_class1_label = "class1 train"
        self.test_class0_label = "class0 test"
        self.test_class1_label = "class1 test"
        self.test_alpha = 0.6
        self.train_alpha = 0.9
        self.x_label = "X axis"
        self.y_label = "Y axis"
        self.color_dict = res.Resources.get_color_dict()





    # Method to set the properties of the graph................................
    # Not necessary, because all the default values are already set......................................

    def set_plot_properties(self, x_label = "X axis", y_label = "Y axis", train_alpha = 0.9,test_alpha = 0.6,test_class1_label = "class1 test",test_class0_label = "class0 test",train_class1_label = "class1 train", train_class0_label = "class0 train", background_color = (0, 0, 0),title = "Logistic regression visulisation",train_class0_color = "#800080", train_class1_color = "#FFA500", test_class0_color = "#00BFFF", test_class1_color = "#FF1493"):
        self.train_class0_color = train_class0_color
        self.train_class1_color = train_class1_color
        self.test_class0_color = test_class0_color
        self.test_class1_color = test_class1_color
        self.title = title
        self.background_color = background_color
        self.train_class0_label = train_class0_label
        self.train_class1_label = train_class1_label
        self.test_class0_label = test_class0_label
        self.test_class1_label = test_class1_label
        self.test_alpha = test_alpha
        self.train_alpha = train_alpha
        self.x_label = x_label
        self.y_label = y_label






    # method that does all the calculatons before plotting the graph.....................................................

    def plot_2D_calculations(self, model1, x, y):
        tsne_data = model1.fit_transform(x)
        class1 = []
        class2 = []

        for i in range(len(y)):
            if y[i] == 0:
                class1.append(tsne_data[i])
            else:
                class2.append(tsne_data[i])

        class1 = np.array(class1).T
        class2 = np.array(class2).T
        return class1, class2






    #Method to plot the 2d graph of the calculated values after apllying T-sne.......................................


    def plot_2D_visuals(self, plot_train_data =True, plot_test_data = False, save_fig_path="dont"):
        model1 = TSNE(n_components=2)

        if plot_train_data:
             class1, class2 = self.plot_2D_calculations(model1, self.model.x, self.model.y)
        if plot_test_data:
             class3, class4 = self.plot_2D_calculations(model1, self.model.x_calculated, self.model.y_calculated)

        plt.style.use('dark_background')

        fig = plt.figure(figsize=(8, 6))

        plt.title(self.title,
                  fontdict=res.Resources.title_dict)


        if plot_train_data:

            plt.scatter(class1[0], class1[1], color=self.train_class0_color, alpha=self.train_alpha, label = self.train_class0_label)
            plt.scatter(class2[0], class2[1], color=self.train_class1_color, alpha=self.train_alpha, label= self.train_class1_label)
        if plot_test_data:

            plt.scatter(class3[0], class3[1], color=self.test_class0_color, alpha=self.test_alpha, label=self.test_class0_label)
            plt.scatter(class4[0], class4[1], color=self.test_class1_color, alpha=self.test_alpha, label=self.test_class1_label)
        plt.legend()
        plt.xlabel(self.x_label,
                   fontdict=res.Resources.label_dict)
        plt.ylabel(self.y_label,
                   fontdict=res.Resources.label_dict)
        if not(save_fig_path=="dont"):
            plt.savefig(save_fig_path + "/2D" + self.title)

        plt.show()
















class Knn_plot():



    #Init method to set default values..................................................

    def __init__(self, model):
        self.model = model
        self.title = "K-nearest neighbours visulisation"
        self.background_color = (0, 0, 0)
        self.test_alpha = 0.6
        self.train_alpha = 0.9
        self.color_dict = res.Resources.get_color_dict()
        self.train_data_color = [key for key in sorted(self.color_dict)[:15]]
        self.test_data_color = [key for key in sorted(self.color_dict)[15:]]
        random.shuffle(self.train_data_color)
        random.shuffle(self.test_data_color)




    # Method to set the colors for train and test data.........................

    def set_colors(self, train_data_color, test_data_color):
        self.train_data_color = train_data_color
        self.test_data_color = test_data_color



    #Method to set the properties of the graph................................
    #Not necessary, because all the default values are already set......................................

    def set_plot_properties(self, train_alpha = 0.9,test_alpha = 0.6, background_color = (0, 0, 0),title = "K-nearest neighbours visulisation"):
        self.title = title
        self.background_color = background_color
        self.test_alpha = test_alpha
        self.train_alpha = train_alpha





    #method that does all the calculatons before plotting the graph.....................................................

    def plot_3D_calculations(self, model1, x, y):
        tsne_data = model1.fit_transform(x)
        classes_dict = {}

        for i in range(len(tsne_data)):
            if y[i] in classes_dict:
                classes_dict[y[i]].append(tsne_data[i])
            else:
                classes_dict[y[i]] = [tsne_data[i]]
        return classes_dict








    # Method to plot the 2d graph of the calculated values after apllying T-sne.......................................

    def plot_3D_visuals(self, plot_train_data =True, plot_test_data = True, save_fig_path="dont"):
        model1 = TSNE(n_components=3)

        if plot_train_data:
             train_classes_dict = self.plot_3D_calculations(model1, self.model.x_data, self.model.y_data)
        if plot_test_data:
             test_classes_dict = self.plot_3D_calculations(model1, self.model.x_classified, self.model.y_classified)

        plt.style.use('dark_background')

        fig = plt.figure(figsize=(12, 12))
        ax = plt.axes(projection='3d')
        ax.text2D(0.30, 0.98, self.title, transform=ax.transAxes)

        ax.set_facecolor(self.background_color)
        if plot_train_data:
            label_dist = 0.02
            ax.text2D(0.86, 0.98, "Train ", transform=ax.transAxes)
            for group in train_classes_dict:
                plotx = []
                ploty = []
                plotz = []
                for coordp in train_classes_dict[group]:
                    plotx.append(coordp[0])
                    ploty.append(coordp[1])
                    plotz.append(coordp[2])

                ax.text2D(0.86, 0.98-label_dist, group, transform=ax.transAxes, color=self.train_data_color[int(label_dist/0.02)%15-1])
                ax.scatter3D(plotx, ploty, plotz, alpha=self.train_alpha, color=self.train_data_color[int(label_dist/0.02)%15-1])
                label_dist+=0.02
        if plot_test_data:
            label_dist = 0.02
            ax.text2D(0.98, 0.98, "Test ", transform=ax.transAxes)
            for group in test_classes_dict:
                plotx = []
                ploty = []
                plotz = []
                for coordp in test_classes_dict[group]:
                    plotx.append(coordp[0])
                    ploty.append(coordp[1])
                    plotz.append(coordp[2])

                ax.text2D(0.98, 0.98-label_dist, group , transform=ax.transAxes, color=self.test_data_color[int(label_dist/0.02)%15-1])
                ax.scatter3D(plotx, ploty, plotz, alpha=self.train_alpha, color=self.test_data_color[int(label_dist/0.02)%15-1])
                label_dist+=0.02
        if not(save_fig_path=="dont"):
            plt.savefig(save_fig_path + "/3D" + self.title)

        plt.show()














class Knn_plot_2D():



    #Init method to set default values..................................................

    def __init__(self, model):
        self.model = model
        self.title = "K-nearest neighbours visulisation"
        self.background_color = (0, 0, 0)
        self.test_alpha = 0.6
        self.train_alpha = 0.9
        self.color_dict = res.Resources.get_color_dict()
        self.train_data_color = [key for key in sorted(self.color_dict)[:15]]
        self.test_data_color = [key for key in sorted(self.color_dict)[15:]]
        random.shuffle(self.train_data_color)
        random.shuffle(self.test_data_color)




    #Method to set the colors for train and test data.........................

    def set_colors(self, train_data_color, test_data_color):
        self.train_data_color = train_data_color
        self.test_data_color = test_data_color



    #Method to set the properties of the graph................................
    #Not necessary, because all the default values are already set......................................

    def set_plot_properties(self, train_alpha = 0.9,test_alpha = 0.6, background_color = (0, 0, 0),title = "K-nearest neighbours visulisation"):
        self.title = title
        self.background_color = background_color
        self.test_alpha = test_alpha
        self.train_alpha = train_alpha





    #method that does all the calculatons before plotting the graph.....................................................

    def plot_3D_calculations(self, model1, x, y):
        tsne_data = model1.fit_transform(x)
        classes_dict = {}

        for i in range(len(tsne_data)):
            if y[i] in classes_dict:
                classes_dict[y[i]].append(tsne_data[i])
            else:
                classes_dict[y[i]] = [tsne_data[i]]
        return classes_dict








    # Method to plot the 2d graph of the calculated values after apllying T-sne.......................................

    def plot_3D_visuals(self, plot_train_data =True, plot_test_data = True, save_fig_path="dont"):
        model1 = TSNE(n_components=2)

        if plot_train_data:
             train_classes_dict = self.plot_3D_calculations(model1, self.model.x_data, self.model.y_data)
        if plot_test_data:
             test_classes_dict = self.plot_3D_calculations(model1, self.model.x_classified, self.model.y_classified)

        plt.style.use('dark_background')

        fig = plt.figure(figsize=(12, 12))
        #ax = plt.axes(projection='3d')
        #ax.text2D(0.30, 0.98, self.title, transform=ax.transAxes)
        plt.title(self.title)

        #ax.set_facecolor(self.background_color)
        if plot_train_data:
            label_dist = 0.02
            #ax.text2D(0.86, 0.98, "Train ", transform=ax.transAxes)
            for group in train_classes_dict:
                plotx = []
                ploty = []
                plotz = []
                for coordp in train_classes_dict[group]:
                    plotx.append(coordp[0])
                    ploty.append(coordp[1])


                #ax.text2D(0.86, 0.98-label_dist, group, transform=ax.transAxes, color=self.train_data_color[int(label_dist/0.02)%15-1])

                plt.scatter(plotx, ploty, alpha=self.train_alpha, label="Train "+str(group), color=self.train_data_color[int(label_dist/0.02)%15-1])
                label_dist+=0.02
        if plot_test_data:
            label_dist = 0.02
            #ax.text2D(0.98, 0.98, "Test ", transform=ax.transAxes)
            for group in test_classes_dict:
                plotx = []
                ploty = []

                for coordp in test_classes_dict[group]:
                    plotx.append(coordp[0])
                    ploty.append(coordp[1])



                #ax.text2D(0.98, 0.98-label_dist, group , transform=ax.transAxes, color=self.test_data_color[int(label_dist/0.02)%15-1])
                plt.scatter(plotx, ploty, alpha=self.train_alpha, label="Test "+str(group),color=self.test_data_color[int(label_dist/0.02)%15-1])
                label_dist+=0.02
        plt.legend()
        if not(save_fig_path=="dont"):
            plt.savefig(save_fig_path + "/3D" + self.title)

        plt.show()














"""

Parallel plot class plots parallel plot graph 

Attributes for the plot can be set in the init method. Otherwise default values are already set.
You just need to pass the x and y to plot the parallel plot.


"""




class plot_parallel_plot():


    #Init mehod to set the default values for the attributes............

    def __init__(self, x, y,
                 save_fig_path='dont',
                 title="Parallel Plot of Dataset",
                 x_label="Parameters/Dimensions",
                 y_label="Normalized Values",
                 alpha=0.5,
                 class_colors=['default'],
                 plot_labels=['default']):


        self.y = np.array(y)
        self.x = np.array(x)
        self.x_normalized = []
        self.plot_labels = []
        self.class_color =[]
        self.x_label = x_label
        self.y_label = y_label
        self.title = title
        self.aplha = alpha
        self.color_dict = res.Resources.get_color_dict()
        self.scatter_color = [key for key in sorted(self.color_dict)]
        self.colorList_len = len(self.scatter_color)
        random.shuffle(self.scatter_color)
        if plot_labels[0] == 'default':
            self.plot_labels = [str(key) for key in np.unique(y)]
        else:
            self.plot_labels = plot_labels

        if class_colors[0] == 'default':
            pass
        else:
            self.class_color = class_colors

        self.plot_plot(self.x, self.y, save_fig_path)







    #Plots the parallel plot. It takes respective x and y as input and also the path to save figure.........

    def plot_plot(self, x, y, save_fig_path='dont'):
            y_classes = np.unique(y)
            y_class_count = len(y_classes)
            self.x_normalized = self.normalize(x)

            plt.style.use('dark_background')
            x_pos = 1
            index = 0
            temp_feature = {}#[[]]*y_class_count
            #print(temp_feature)
            for feature in self.x_normalized:

                temp_feature = {}# [[]] * y_class_count
                for i in range(len(feature)):
                    if y[i] in temp_feature:
                        temp_feature[y[i]].append(feature[i])
                    else:
                        temp_feature[y[i]] = [feature[i]]
                index = 0
                for key in temp_feature:

                    if len(self.class_color) == 0:
                        plt.scatter([x_pos]*len(temp_feature[key]), temp_feature[key], alpha=self.aplha,color=self.scatter_color[index%self.colorList_len])
                    else:
                        plt.scatter([x_pos] * len(temp_feature[key]), temp_feature[key], alpha=self.aplha,
                                    color=self.class_color[index % len(self.class_color)])

                    index += 1
                x_pos += 1
                #print(temp_feature)

            index=0
            for key in temp_feature:

                if len(self.plot_labels) == 0 and len(self.class_color)==0:
                    plt.scatter([], [], label="class "+str(key), color=self.scatter_color[index%self.colorList_len])
                elif len(self.plot_labels) == 0:
                    plt.scatter([], [], label="class "+str(key), color=self.class_color[index % len(self.class_color)])

                elif len(self.class_color) == 0 :
                    plt.scatter([], [], label=str(self.plot_labels[index % len(self.plot_labels)]), color=self.scatter_color[index % self.colorList_len])

                else:
                    plt.scatter([], [], label=str(self.plot_labels[index%len(self.plot_labels)]), color=self.class_color[index % len(self.class_color)])
                index+=1

            plt.ylabel(self.y_label)
            plt.xlabel(self.x_label)
            plt.title(self.title)
            plt.legend()
            if not(save_fig_path=="dont"):
                plt.savefig(save_fig_path + "/3D" + self.title)
            plt.show()






    # Normalize the values of the parameters to let them lie between 0 and 1

    def normalize(self, x):
        x = x.T
        temp_sum = 0
        x_normalized = []
        for i, feature in enumerate(x):
             x_normalized.append((feature/np.sum(feature)))
             #print(list(x_normalized[i]))

        return x_normalized




















class ninja_technique():

    def __init__(self):

        self.x_coords=[]
        self.y_oords=[]
        self.x_coords1 = []
        self.y_oords1 = []
        self.normalize_x = []
        self.tita = 0

    def calculate_normalisation(self, x):

        self.normalize_x = [0 for x in range(len(x[0]))]

        for i in x:
            for index,j in enumerate(i):
                self.normalize_x[index] += j

        #print(self.normalize_x)




    def calculate(self,model, x_data, y_data):
        self.tita = 90/(len(model.weights)-1)
        self.calculate_normalisation(x_data)
        print(self.tita, len(model.weights))

        for index,i in enumerate(x_data):
            temp_x = 0#model.bias
            temp_y = 0#model.bias
            temp_tita = 0
            for j in range(len(i)):
                temp_x = temp_x + (model.weights[j]*i[j]*math.cos(math.radians(temp_tita)))/self.normalize_x[j]
                temp_y = temp_y + (model.weights[j]*i[j]*math.sin(math.radians(temp_tita)))/self.normalize_x[j]

                #temp_x = temp_x + (i[j] * math.cos(math.radians(temp_tita)))/self.normalize_x[j]
                #temp_y = temp_y + (i[j] * math.sin(math.radians(temp_tita)))/self.normalize_x[j]

                temp_tita+=self.tita


            if y_data[index] == 0:
                self.x_coords.append(temp_x)
                self.y_oords.append(temp_y)
            else:
                self.x_coords1.append(temp_x)
                self.y_oords1.append(temp_y)







    def plot(self):
        plt.style.use('dark_background')
        plt.title("Swiss roll dataset plot")
        plt.scatter(self.x_coords, self.y_oords, label="class 0", color="#800080", alpha=0.3)
        plt.scatter(self.x_coords1, self.y_oords1, label="class 1", color='#FF8C00', alpha=0.3)
        plt.legend()
        plt.show()

    def plot_multi(self ,x1, x2, x3):
        plt.style.use('dark_background')
        plt.title("Wine dataset plot")
        plt.scatter(x1[0], x1[1], label="class 0", color="#800080", alpha=0.3)
        plt.scatter(x2[0], x2[1], label="class 1", color='#FF8C00', alpha=0.3)
        plt.scatter(x3[0], x3[1], label="class 2", color='#00FFFF', alpha=0.3)
        plt.legend()
        plt.show()
















"""
bc = datasets.load_breast_cancer()
x, y = bc.data, bc.target

x_train, x_test, y_train, y_test = train_test_split(x, y, test_size=0.2, random_state=1234)


lor = lore.LogisticRegression(learning_rate=0.01, n_iters=10000)
lor.best_fit(x_train, y_train)
lor.best_fit(x, y)
y_pred = lor.predict(x)
prediction = lor.predict(x_test)

accuracy = np.sum(prediction == y_test)/ len(prediction)
print(prediction)
print(list(y_test))
print(accuracy)



plotly = ninja_technique()
plotly1 = ninja_technique()
#plotly.calculate(lor, x_train + x_test, y_train + y_test)
plotly.calculate(lor, x, y_pred)
plotly.plot()
#plotly1.calculate(lor, x_test, y_test)
#plotly1.plot()


datav = Logistic_regression_plot2D(lor)
datav.set_plot_properties(train_class0_label="benign", train_class1_label="malignant")
datav.plot_2D_visuals(plot_test_data=False)



"""



"""




data = sklearn.datasets.make_swiss_roll(n_samples=5000, noise=0.2, random_state=None)
ldata_x = data[0]
ldata_y = []
#print(data)
#for i in data:
    #print(i)
data = np.array(data[0]).T
plt.style.use('dark_background')
#print(data)

data1 = [[], [], []]
data2 = [[], [], []]

error = [-4, -3, -5]

for i in range(len(data[0])):
    if data[0][i] <= (7 ) and (data[2][i]>-6+ random.random()*random.choice(error) and data[2][i]<10):#- random.random()*random.choice(error)):
        data1[0].append(data[0][i])
        data1[1].append(data[1][i])
        data1[2].append(data[2][i])
        ldata_y.append(0)
    else:
        data2[0].append(data[0][i])
        data2[1].append(data[1][i])
        data2[2].append(data[2][i])
        ldata_y.append(1)



lor1 = lore.LogisticRegression(learning_rate=0.03,n_iters=100000)
lor1.best_fit(ldata_x, ldata_y)
y_data = lor1.predict(ldata_x)
count = 0

for i in range(len(ldata_y)):
    if ldata_y[i] == y_data[i]:
        count+=1

new_data = data1 + data2
data11= []
data12 = []




print("Accuracy = " + str(count/len(ldata_y) * 100))
#plot = Logistic_regression_plot2D(lor1)
#plot.set_plot_properties(title="Swiss Roll dataset")
#plot.plot_2D_visuals()
nj = ninja_technique()
nj.calculate(lor1, ldata_x, ldata_y)
nj.plot()



print(data1[0])
print(data1[1])
print(data1[2])
print(data2)




fig = plt.figure(figsize=(12, 12))
ax = plt.axes(projection='3d')
ax.scatter3D(data1[0], data1[1], data1[2], color="#FF4500", alpha=0.6)
ax.scatter3D(data2[0], data2[1], data2[2], color="#800080", alpha=0.6)
plt.show()

"""






"""

data = sklearn.datasets.make_swiss_roll(n_samples=5000, noise=0.0, random_state=None)
ldata_x = []
ldata_y = []
#print(data)
#for i in data:
    #print(i)
data = np.array(data[0]).T
plt.style.use('dark_background')
#print(data)

data1 = [[], [], []]
data2 = [[], [], []]

error = [-4, -3, -5]

for i in range(len(data[0])):

    if data[0][i] > 2:
        ldata_x.append([data[0][i], data[1][i], data[2][i]])

        if data[0][i] <= (7 ) and (data[2][i]>-6+ random.random()*random.choice(error) and data[2][i]<10):#- random.random()*random.choice(error)):
            data1[0].append(data[0][i])
            data1[1].append(data[1][i])
            data1[2].append(data[2][i])
            ldata_y.append(0)
        else:
            data2[0].append(data[0][i])
            data2[1].append(data[1][i])
            data2[2].append(data[2][i])
            ldata_y.append(1)




lor1 = lore.LogisticRegression(n_iters=100000)
lor1.best_fit(ldata_x, ldata_y)
y_data = lor1.predict(ldata_x)
count = 0
for i in range(len(ldata_y)):
    if ldata_y[i] == y_data[i]:
        count+=1


print("Accuracy = " + str(count/len(ldata_y) * 100))
plot = Logistic_regression_plot2D(lor1)
plot.set_plot_properties(title="Swiss Roll dataset")
plot.plot_2D_visuals()
nj = ninja_technique()
nj.calculate(lor1, ldata_x, ldata_y)
nj.plot()




fig = plt.figure(figsize=(12, 12))
ax = plt.axes(projection='3d')
ax.scatter3D(data1[0], data1[1], data1[2], color="#800080", alpha=0.6)
ax.scatter3D(data2[0], data2[1], data2[2], color="#FF4500", alpha=0.6)
plt.show()

"""











"""
data = sklearn.datasets.make_s_curve(n_samples=10000, noise=0.0, random_state=None)

data1 = [[], [], []]
data2 = [[], [], []]
y_values = []
x_data = []
#data = np.array(data[0]).T
#print(data[0])
#error = [-0.25, 0.25]
error = [0, 0]
for i in data[0]:
    if (-1.1<i[2] and i[2]<1.1 and i[0]+i[2]<1.5 and i[0]+i[2]>-1.5 ):
        x_data.append(i)
        if i[0] > 0 + random.random()*random.choice(error):
            data1[0].append(i[2]*100)
            data1[1].append(i[1]*100)
            data1[2].append(i[0]*100)
            y_values.append(1)
        else:
            data2[0].append(i[2]*100)
            data2[1].append(i[1]*100)
            data2[2].append(i[0]*100)
            y_values.append(0)
data1 = np.array(data1)
data2 = np.array(data2)

lor1 = lore.LogisticRegression(learning_rate=0.3, n_iters=100000)
lor1.best_fit(x_data, y_values)
y_data = lor1.predict(x_data, decision_boundary=0.1)
count = 0
print(len(y_values))
print(y_data)
for i in range(len(y_values)):
    if y_values[i] == y_data[i]:
        count+=1

print("Accuracy = "+ str(count*100/len(y_values)))



plot = Logistic_regression_plot2D(lor1)
plot.set_plot_properties(title="Swiss Roll dataset")
plot.plot_2D_visuals()

 ninja_technique()
nj.calculate(lor1, x_data, y_data)
nj.plot()


plt.style.use('dark_background')
fig = plt.figure(figsize=(12, 12))
ax = plt.axes(projection='3d')
ax.scatter3D(data1[0], data1[1], data1[2], alpha=0.6, color="#800080")
ax.scatter3D(data2[0], data2[1], data2[2],alpha=0.6, color="#FF4500")

plt.show()


"""





"""

bc = datasets.load_iris()
x, y = bc.data, bc.target

x_train, x_test, y_train, y_test = train_test_split(x, y, test_size=0.2, random_state=1234)

#print(x)
print(y)


lr1 = sklearn.linear_model.LogisticRegression(max_iter=1000000)
lr1.fit(x, y)



y1 = lr1.predict(x)
count = 0
for i, j in enumerate(y):
    if j == y1[i]:
        count+=1
    else:
        print("Class " + str(j) + " " + str(y1[i]))
print(y1)
print("Accuracy = " + str(count/len(y) * 100))

x1 = [[], []]
x2 = [[], []]
x3 = [[], []]

weight = lr1.coef_
normalize = [0, 0, 0, 0]
print(weight)

for i in x:
    normalize[0] += i[0]
    normalize[1] += i[1]
    normalize[2] += i[2]
    normalize[3] += i[3]

for i, j in enumerate(x):

    alpha = 90/(len(x[0])-1)
    temp_y =0
    temp_x = 0
    temp_tita = 0
    for k in range(len(j)):
        temp_x = temp_x + (weight[int(y[i])][k] * j[k] * math.cos(math.radians(temp_tita))) / normalize[k]
        temp_y = temp_y + (weight[int(y[i])][k] * j[k] * math.sin(math.radians(temp_tita))) / normalize[k]
        #temp_x = temp_x + (j[k] * math.cos(math.radians(temp_tita))) / normalize[k]
        #temp_y = temp_y + (j[k] * math.sin(math.radians(temp_tita))) / normalize[k]
        temp_tita +=alpha

    if y1[i] == 0:
        x1[0].append(temp_x)
        x1[1].append(temp_y)
    elif y1[i] ==1:
        x2[0].append(temp_x)
        x2[1].append(temp_y)
    else:
        x3[0].append(temp_x)
        x3[1].append(temp_y)

print(x1,x2,x3)
nj = ninja_technique()
nj.plot_multi(x1, x2, x3)


print(lr1.coef_)



model1 = TSNE(n_components=2)

tsne_data = model1.fit_transform(x)
classes_dict = {}

for i in range(len(tsne_data)):
    if y[i] in classes_dict:
        classes_dict[y[i]].append(tsne_data[i])
    else:
        classes_dict[y[i]] = [tsne_data[i]]

print(classes_dict.keys())

x11 = np.array(classes_dict[0]).T
x21 = np.array(classes_dict[1]).T
x31 = np.array(classes_dict[2]).T

nj.plot_multi(x11, x21, x31)
"""










"""
bc = datasets.load_wine()
x, y = bc.data, bc.target

print(x)
print(y)


lr1 = sklearn.linear_model.LogisticRegression(max_iter=1000000)
lr1.fit(x, y)



y1 = lr1.predict(x)
count = 0
for i, j in enumerate(y):
    if j == y1[i]:
        count+=1
    else:
        print("Class " + str(j) + " " + str(y1[i]))
print(y1)
print("Accuracy = " + str(count/len(y) * 100))

x1 = [[], []]
x2 = [[], []]
x3 = [[], []]

weight = lr1.coef_
normalize = [0 for i in range(len(x[0]))]
print(weight)

for i in x:
    for j in range(len(i)):
        normalize[j] += i[j]


for i, j in enumerate(x):

    alpha = 90/(len(x[0])-1)
    temp_y =0
    temp_x = 0
    temp_tita = 0
    for k in range(len(j)):
        temp_x = temp_x + (weight[int(y[i])][k] * j[k] * math.cos(math.radians(temp_tita))) / normalize[k]
        temp_y = temp_y + (weight[int(y[i])][k] * j[k] * math.sin(math.radians(temp_tita))) / normalize[k]
        #temp_x = temp_x + (j[k] * math.cos(math.radians(temp_tita))) / normalize[k]
        #temp_y = temp_y + (j[k] * math.sin(math.radians(temp_tita))) / normalize[k]
        temp_tita +=alpha

    if y[i] == 0:
        x1[0].append(temp_x)
        x1[1].append(temp_y)
    elif y[i] ==1:
        x2[0].append(temp_x)
        x2[1].append(temp_y)
    else:
        x3[0].append(temp_x)
        x3[1].append(temp_y)

print(x1,x2,x3)
nj = ninja_technique()
nj.plot_multi(x1, x2, x3)


print(lr1.coef_)



model1 = TSNE(n_components=2)

tsne_data = model1.fit_transform(x)
classes_dict = {}

for i in range(len(tsne_data)):
    if y[i] in classes_dict:
        classes_dict[y[i]].append(tsne_data[i])
    else:
        classes_dict[y[i]] = [tsne_data[i]]

print(classes_dict.keys())

x11 = np.array(classes_dict[0]).T
x21 = np.array(classes_dict[1]).T
x31 = np.array(classes_dict[2]).T

nj.plot_multi(x11, x21, x31)


"""

"""

bc = datasets.load_breast_cancer()
x, y = bc.data, bc.target


plt1 = plot_parallel_plot(x, y, class_colors=['Red', 'Cyan', 'Yellow'], alpha=0.2)
"""





"""



This file implements Naive Bayes algorithm. This algorithm is based on bayes therom.

This file has class that  can also visualise it.

This is a preliminary file that performs Naive Bayes. This is one of the first versions. So a lot
will be improved in the future.


Classes:
Naive Bayes- It carries out model
plot model- It carries out data visualisation.





Attributes:
............................
____________________________
x                  -> input x parameters (list of list or np array)
y                  -> input y for training data
showGrid           -> Boolean which indicates weather grid should be shown or not
sizeX              -> Horizontal size of graph in inches
sizeY              -> Vertical size of graph in inches
labelx_count       -> number of labels on x axis
labely_count       -> number of labels on y axis
title_color        -> color of title of plot
label_default      -> Boolean which tells if the label is default or changed by user
xlabel_color       -> color of x label
ylabel_color       -> color of y label
title              -> title text
x_label            -> List x label text for all parameters/dimentions
y_label            -> y label text
xlabel_size        -> x_label font size of text
ylabel_size        -> y_label font size of text
title_size         -> title font size of text
model              -> this is the model object i.e. Naive bayes object in case we need any parent class variables
x_calculated       -> The calculated part of model's x parameters or the test data's x coordinates
y_calculated       -> Calculated classification using model or the test data
learning_rate      -> The rate at with we want to train our model (usually in order of 0.001)
n_iters            -> Number of iterations which we we want our model to pass through the training process
weights            -> List of weights used in our model
bias               -> single float/double containing our bias value
parameters         -> count of x parameters, dimentions
coordinates_size   -> size of training data
calculated_point_class0_label -> train data class 0 color
calculated_point_class1_label -> train data class 1 point color
calculated_point_class0_color -> test data class 0 point color
calculated_point_class1_color -> test data class 1 point color
calculated_point_alpha        -> alpha/transparency of the point


#####.....IMP.......####
display_graph -> Specify weather graph needs to be displayed before saving in plot_graph method
color_dict -> Dictionary that contains wide range of colors in the form of dictionary. The keys are the names of colors
              , values are the hex codes of that color. You can get this dict using get_color_dict function()
You can print this dict i.e. print(object.get_color_dict()) to know which colors are available.






You are most welcomed to improve this code:
You can pull request for this code at

github.com/mrfrozen97/                      (In spiderAlgorithms repo)
or
Email - mrfrozenpeak@gmail.com




"""



import numpy as np
#from sklearn.model_selection import train_test_split
#from sklearn import datasets
import matplotlib.pyplot as plt
import random
from frozenSpider import spiderAlgorithmResources as res












class NaiveBayes:

    def __init__(self):
        self.x_data = []
        self.y_data = []
        self.x_classified = []
        self.y_classified = []




    def fit(self, x, y):

        self.x_data =x
        self.y_data =y

        n_samples, n_features = x.shape
        self.y_claasses = np.unique(y)
        self.y_class_count = len(self.y_claasses)


        self.x_means = np.zeros((self.y_class_count, n_features), dtype=np.float64)
        self.x_vars = np.zeros((self.y_class_count, n_features), dtype=np.float64)
        self.priors = np.zeros((self.y_class_count), dtype=np.float64)

        for c in self.y_claasses:
            x_c = x[c==y]
            self.x_means[c, :] = x_c.mean(axis=0)
            self.x_vars[c, :] = x_c.var(axis=0)
            self.priors[c] = x_c.shape[0] / float(n_samples)


    def predict(self, x):
        y_predict = [self.predict_sample(i) for i in x]
        self.x_classified = x
        self.y_classified = y_predict

        return y_predict


    def predict_sample(self, x):

        probabilities_calculated = []

        for i, c in enumerate(self.y_claasses):
            prior = np.log(self.priors[i])
            class_conditional = np.sum(np.log(self.gaussian_normal_function(i, x)))
            probabilities_calculated.append(prior+class_conditional)

        return self.y_claasses[np.argmax(probabilities_calculated)]


    def gaussian_normal_function(self, class_index, x):
        mean = self.x_means[class_index]
        var = self.x_vars[class_index]
        numerator = np.exp(-(x-mean)**2 / (2*var))
        denominator = np.sqrt(2*np.pi * var)
        return numerator/denominator















class plot_model():

    def __init__(self, model):
        self.showGrid = False
        self.sizeX = 10
        self.sizeY = 6
        self.labelx_count = 10
        self.labely_count = 10
        self.title_color = "#FF0000"
        self.xlabel_color = "#663399"
        self.ylabel_color = "#663399"
        self.label_default = True
        self.title = ""
        self.x_label = ["Dimention " + str(x) for x in range(1, 100)]
        self.y_label = ""
        self.xlabel_size = 15
        self.ylabel_size = 15
        self.title_size = 18
        self.model = model
        self.color_dict = res.Resources.get_color_dict()
        self.calculated_point_size = 10
        self.calculated_point_alpha = 0.5
        self.train_labels = []
        self.test_labels = []
        self.train_data_color = [key for key in sorted(self.color_dict)[:15]]
        self.test_data_color = [key for key in sorted(self.color_dict)[15:]]
        random.shuffle(self.train_data_color)
        random.shuffle(self.test_data_color)
        self.legend_position = "upper right"
        self.plot_background = 'dark'








    # Function to get dictionary to access the different options avaliable for the colors.........................

    def get_color_dict(self):
        return self.color_dict








    # function to set label of each dimention of the graph................................................

    def set_dimention_labels(self, dimention_label):
        self.x_label = dimention_label







    def set_marker_properties(self,show_grid = False, plot_background='dark', legend_position="upper right", train_labels=[], test_labels=[], label_default=True, calculated_point_class1_label = "Calculated class 1", calculated_point_class0_label="Calculated class 0", calculated_point_alpha = 0.5,  calculated_point_class1_color='Red',calculated_point_class0_color="Cyan", title_size=15, xlabel_size=10, ylabel_size=10, title="Output vs dimention", y_label="y coordinates", title_color="#FF0000", xlabel_color="#663399", ylabel_color="#663399"):
        self.title_size = title_size
        self.xlabel_size = xlabel_size
        self.ylabel_size = ylabel_size
        self.y_label = y_label
        self.title = title
        self.showGrid = show_grid

        self.xlabel_color = xlabel_color
        self.ylabel_color = ylabel_color
        self.title_color = title_color
        self.label_default = label_default
        self.calculated_point_size = 10

        self.calculated_point_class0_label = calculated_point_class0_label
        self.calculated_point_class1_label = calculated_point_class1_label
        self.calculated_point_class0_color = calculated_point_class0_color
        self.calculated_point_class1_color = calculated_point_class1_color

        self.calculated_point_alpha = calculated_point_alpha
        self.train_labels = train_labels
        self.test_labels = test_labels
        self.legend_position = legend_position
        self.plot_background = plot_background







  #Sets the display size which will be equal to size of image if it is being svaed.........................

    def set_display_size(self, sizeX, sizeY):
        self.sizeX = sizeX
        self.sizeY = sizeY





    # main fuction thats plots the graph and saved it if path provided.........................
    #This function iterates through all the dimentions/parameters and plot each one seperate...............




    def plot_model(self, display_graph = True,display_calculated_points = True, alpha=0.6, point_size=25,label_x=[], label_y=[],class1_color='Purple', class0_color='Orange',unknown_points_label="Unknown points",line_label='Best Fit line', point_color='DeepSkyBlue', save_fig_file='dont'):


        # These are the coordinates to plot the line which is inclusion of both known and calculated points
        if self.plot_background=='dark':
              plt.style.use('dark_background')
        multi_dimentions = np.array(self.model.x_data)
        multi_dimentions_calculated = np.array(self.model.x_classified)
       # print(len(multi_dimentions[0]))
        #print(multi_dimentions)

        for dimention in range(len(multi_dimentions[0])):

            #x_train_data = []
            y_train_data = {}
            y_test_data = {}

            for i,a in enumerate(multi_dimentions):
                #print(i)
                if self.model.y_data[i] in y_train_data:
                    y_train_data[self.model.y_data[i]].append(a[dimention])
                else:
                    y_train_data[self.model.y_data[i]] = [a[dimention]]

            for i, a in enumerate(multi_dimentions_calculated):

                if self.model.y_classified[i] in y_test_data:
                    y_test_data[self.model.y_classified[i]].append(a[dimention])
                else:
                    y_test_data[self.model.y_classified[i]] = [a[dimention]]
                #x_train_data.append(i[dimention])
            #print(y_test_data)

           # print(y_train_data)

            index_color = 0
            for group in y_train_data:
                #print(group)
                y_plot_axis = []

                if len(self.train_labels)==0:
                    actual_train_label = "train class" + str(group)
                else:
                    actual_train_label = self.train_labels[index_color]

                for i in range(len(y_train_data[group])):
                    y_plot_axis.append(group)
                plt.scatter(y_train_data[group], y_plot_axis,
                            color=self.color_dict[self.train_data_color[index_color%15]],
                            label=actual_train_label,
                            alpha=alpha,
                            zorder=3)                     #plot the best fit line
                index_color+=1


            index_color = 0

            if display_calculated_points:

                for group in y_test_data:
                    #print(group)
                    y_plot_axis = []

                    if len(self.test_labels) == 0:
                        actual_test_label = "test class" + str(group)
                    else:
                        actual_test_label = self.test_labels[index_color]


                    for i in range(len(y_test_data[group])):
                        y_plot_axis.append(group)

                    plt.scatter(y_test_data[group], y_plot_axis,
                                label=actual_test_label,
                                alpha=alpha,
                                zorder=3,
                                color=self.color_dict[self.test_data_color[index_color%15]])  # plot the best fit line

                    index_color +=1



            plt.title(self.title,
                      fontdict=res.Resources.title_dict,
                      color=self.title_color)
            plt.xlabel(self.x_label[dimention],
                       fontdict=res.Resources.label_dict,
                       color=self.xlabel_color)
            plt.ylabel(self.y_label,
                       fontdict=res.Resources.label_dict,
                       color=self.ylabel_color)





            plt.legend(loc = self.legend_position)

            if self.showGrid:
                plt.grid(color='#cfd8dc', zorder=0)

            figure = plt.gcf()
            figure.set_size_inches(self.sizeX, self.sizeY)


            if not(save_fig_file=='dont') :

                  plt.savefig("./"+save_fig_file +"/"+self.x_label[dimention])#, bbox_inches='tight')
            if display_graph:
               plt.show()
            plt.close()












#Example breast cancer dataset
"""
bc = datasets.load_breast_cancer()
x, y = bc.data, bc.target
x_train, x_test, y_train, y_test = train_test_split(x, y, test_size=0.2, random_state=123)


nb = NaiveBayes()
nb.fit(x_train, y_train)
y_predict = nb.predict(x_test)

accuracy = (np.sum(y_test==y_predict)/len(y_test) ) * 100

print("Accuracy = " + str(accuracy))
"""



#Example 2 Wine dataset
"""
bc = datasets.load_wine()
x, y = bc.data, bc.target
x_train, x_test, y_train, y_test = train_test_split(x, y, test_size=0.2, random_state=123)


nb = NaiveBayes()
nb.fit(x_train, y_train)
y_predict = nb.predict(x_test)

plot1 = plot_model(nb)
plot1.plot_model()


accuracy = (np.sum(y_test==y_predict)/len(y_test) ) * 100
print(list(y_predict))
print(list(y_test))

print("Accuracy = " + str(accuracy))


"""




#Example 3 Iris dataset
"""
bc = datasets.load_iris()
x, y = bc.data, bc.target
x_train, x_test, y_train, y_test = train_test_split(x, y, test_size=0.2, random_state=123)


nb = NaiveBayes()
nb.fit(x_train, y_train)
y_predict = nb.predict(x_test)

accuracy = (np.sum(y_test==y_predict)/len(y_test) ) * 100

print("Accuracy = " + str(accuracy))


"""
